import numpy
from enum import IntEnum
from ..lattice import Lattice, AtWarning
from at.lattice import RFCavity, Collective
from at.lattice.elements import _array
from at.lattice.utils import Refpts, uint32_refpts, make_copy
from at.physics import get_timelag_fromU0
from at.constants import clight
from typing import Sequence, Optional, Union
import warnings

class BLMode(IntEnum):
    WAKE = 1
    PHASOR = 2


class CavityMode(IntEnum):
    ACTIVE = 1
    PASSIVE = 2


def add_beamloading(ring: Lattice, qfactor: Union[float, Sequence[float]],
                    rshunt: Union[float, Sequence[float]],
                    cavpts: Refpts = None, copy: Optional[bool] = False,
                    **kwargs):
    r"""Function to add beam loading to a cavity element, the cavity
    element is changed to a beam loading element that combines the energy
    kick from both the cavity and the resonator

    Parameters:
        ring:           Lattice object
        qfactor:        Q factor. Scalar or array of float are accepted
        rshunt:         Shunt impedance, [:math:`\Omega`] for longitudinal
                        Scalar or array of float are accepted

    Keyword Arguments:
        cavpts (Refpts):    refpts of the cavity. If None (default) apply
                            to all cavity in the ring
        Nslice (int):       Number of slices per bunch. Default: 101
        Nturns (int):       Number of turn for the wake field. Default: 1
        ZCuts:              Limits for fixed slicing, default is adaptive
        NormFact (float):   Normalization factor
        blmode (BLMode):  method for beam loading calculation BLMode.PHASOR
            (default) uses the phasor method, BLMode.WAKE uses the wake
            function. For high Q resonator, the phasor method should be
            used
        copy:       If True, returns a shallow copy of ring with new
                    beam loading elements. Otherwise, modify ring in-place
        cavitymode (CavityMode):     Define PASSIVE or ACTIVE cavity
        buffersize (int):  Size of the history buffer for vbeam, vgen, vbunch
                (default 0)
    """
    @make_copy(copy)
    def apply(ring, cavpts, newelems):
        for ref, elem in zip(cavpts, newelems):
            ring[ref] = elem

    if cavpts is None:
        cavpts = ring.get_refpts(RFCavity)
    else:
        cavpts = uint32_refpts(cavpts, len(ring))
    qfactor = numpy.broadcast_to(qfactor, (len(cavpts), ))
    rshunt = numpy.broadcast_to(rshunt, (len(cavpts), ))
    new_elems = []
    for ref, qf, rs in zip(cavpts, qfactor, rshunt):
        cav = ring[ref]
        if not isinstance(cav, RFCavity):
            raise TypeError('Beam loading can only be assigned' +
                            'to an RFCavity element')
        new_elems.append(BeamLoadingElement.build_from_cav(cav, ring, qf,
                                                           rs, **kwargs))
    return apply(ring, cavpts, new_elems)


def remove_beamloading(ring, cavpts: Refpts = None,
                       copy: Optional[bool] = False):
    """Function to remove beam loading from a cavity element, the beam
    loading element is changed to a beam loading element that
    combines the energy kick from both the cavity and the resonator

    Parameters:
        ring:           Lattice object

    Keyword Arguments:
        cavpts (Refpts):    refpts of the beam loading Elements.
                            If None (default) apply to all elements
        copy:       If True, returns a shallow copy of ring with new
                      cavity elements. Otherwise, modify ring in-place
    """
    @make_copy(copy)
    def apply(ring, cavpts, newelems):
        for ref, elem in zip(cavpts, newelems):
            ring[ref] = elem

    if cavpts is None:
        cavpts = ring.get_refpts(BeamLoadingElement)
    else:
        cavpts = uint32_refpts(cavpts, len(ring))
    new_elems = []
    for ref in cavpts:
        bl = ring[ref]
        if not isinstance(bl, BeamLoadingElement):
            raise TypeError('Cannot remove beam loading: ' +
                            'not a BeamLoadingElement')
        family_name = bl.FamName.replace('_BL', '')
        harm = numpy.round(bl.Frequency*ring.circumference/clight)
        new_elems.append(RFCavity(family_name, bl.Length, bl.Voltage,
                                  bl.Frequency, harm, bl.Energy,
                                  TimeLag=getattr(bl, 'TimeLag', 0.0),
                                  PhaseLag=getattr(bl, 'PhaseLag', 0.0)))
    return apply(ring, cavpts, new_elems)


class BeamLoadingElement(RFCavity, Collective):
    """Class to generate a beamloading element, inherits from Element
       additional argument are ring, cavity, qfactor, rshunt
    """
    default_pass = {False: 'DriftPass', True: 'BeamLoadingCavityPass'}
    _conversions = dict(RFCavity._conversions,
                        Rshunt=float, Qfactor=float, NormFact=float,
                        PhaseGain=float, VoltGain=float, _blmode=int,
                        _beta=float, _wakefact=float, _nslice=int,
                        ZCuts=lambda v: _array(v), _cavitymode=int,
                        _nturns=int, _phis=float, _buffersize=int,
                        _vbeam_phasor=lambda v: _array(v, shape=(2,)),
                        _vbeam=lambda v: _array(v, shape=(2,)),
                        _vcav=lambda v: _array(v, shape=(2,)),
                        _vgen=lambda v: _array(v, shape=(2,))
                        )

    def __init__(self, family_name: str, length: float, voltage: float,
                 frequency: float, ring: Lattice, qfactor: float,
                 rshunt: float, blmode: Optional[BLMode] = BLMode.PHASOR,
                 cavitymode: Optional[CavityMode] = CavityMode.ACTIVE,
                 buffersize: Optional[int] = 0, **kwargs):
        r"""
        Parameters:
            ring:            Lattice object
            length:          Length of the cavity
            voltage:         Cavity voltage [V]
            frequency:       Cavity frequency [Hz]
            qfactor:         Q factor
            rshunt:          Shunt impedance, [:math:`\Omega`]

        Keyword Arguments:
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (float):   Normalization factor
            blmode (BLMode):  method for beam loading calculation BLMode.PHASOR
                (default) uses the phasor method, BLMode.WAKE uses the wake
                function. For high Q resonator, the phasor method should be
                used
            cavitymode (CavityMode):  Is cavity ACTIVE (default) or PASSIVE
            buffersize (int):  Size of the history buffer for vbeam, vgen, vbunch
                (default 0)
        Returns:
            bl_elem (Element): beam loading element
        """
        kwargs.setdefault('PassMethod', self.default_pass[True])
        if not isinstance(blmode, BLMode):
            raise TypeError('blmode mode has to be an ' +
                            'instance of BLMode')
        if not isinstance(cavitymode, CavityMode):
            raise TypeError('cavitymode has to be an ' +
                            'instance of CavityMode')
        zcuts = kwargs.pop('ZCuts', None)
        phil = kwargs.pop('phil', 0)
        energy = ring.energy
        harmonic_number = numpy.round(frequency*ring.circumference/clight)
        self.Rshunt = rshunt
        self.Qfactor = qfactor
        self.NormFact = kwargs.pop('NormFact', 1.0)
        self.PhaseGain = kwargs.pop('PhaseGain', 1.0)
        self.VoltGain = kwargs.pop('VoltGain', 1.0)
        self._blmode = int(blmode)
        self._cavitymode = int(cavitymode)
        self._beta = ring.beta
        self._wakefact = - ring.circumference/(clight *
                                               ring.energy*ring.beta**3)
        self._nslice = kwargs.pop('Nslice', 101)
        self._nturns = kwargs.pop('Nturns', 1)
        self._nbunch = ring.nbunch
        self._turnhistory = None    # Defined here to avoid warning
        self._vbunch = None
        self._buffersize = buffersize
        self._vgen_buffer = numpy.zeros(1)
        self._vbeam_buffer = numpy.zeros(1)
        self._vbunch_buffer = numpy.zeros(1)
        if zcuts is not None:
            self.ZCuts = zcuts
        super(BeamLoadingElement, self).__init__(family_name, length,
                                                 voltage, frequency,
                                                 harmonic_number,
                                                 energy, **kwargs)
        _, ts = get_timelag_fromU0(ring)
        self._phis = 2*numpy.pi*self.Frequency*(ts+self.TimeLag)/clight
        self._vbeam_phasor = numpy.zeros(2)
        self._vbeam = numpy.zeros(2)
        self._vgen = numpy.zeros(2)
        self._vcav = numpy.array([self.Voltage,
                                  numpy.pi/2-self._phis-phil])
        self.clear_history(ring=ring)
        
    def is_compatible(self, other):
        return False

    def clear_history(self, ring=None):
        if ring is not None:
            self._nbunch = ring.nbunch
            current = ring.beam_current
            nbunch = ring.nbunch
            self._vbunch = numpy.zeros((nbunch, 2), order='F')
            self._init_bl_params(current)
        tl = self._nturns * self._nslice * self._nbunch
        self._turnhistory = numpy.zeros((tl, 4), order='F')
        if self._buffersize > 0:
            self._vgen_buffer = numpy.zeros((2, self._buffersize),
                                            order='F')
            self._vbeam_buffer = numpy.zeros((2, self._buffersize),
                                             order='F')
            self._vbunch_buffer = numpy.zeros((self._nbunch, 2, self._buffersize),
                                              order='F')

    def _init_bl_params(self, current):
        if (self._cavitymode == 1) and (current > 0.0):
            theta = -self._vcav[1]+numpy.pi/2
            vb = 2*current*self.Rshunt
            a = self.Voltage*numpy.cos(theta-self._phis)
            b = self.Voltage*numpy.sin(theta-self._phis)-vb*numpy.cos(theta)
            psi = numpy.arcsin(b/numpy.sqrt(a**2+b**2))
            vgen = self.Voltage*numpy.cos(psi) + \
                vb*numpy.cos(psi)*numpy.sin(self._phis)
        elif self._cavitymode == 2:
            vgen = 0
            psi = 0
        else:
            vgen = self.Voltage
            psi = 0
            
        self._vbeam = numpy.array([2*current*self.Rshunt*numpy.cos(psi),
                                   numpy.pi-psi])
        self._vgen = numpy.array([vgen, psi])

    @property
    def Buffersize(self):
        return self._buffersize

    @Buffersize.setter
    def Buffersize(self, value):
        self._buffersize = value
        self.clear_history()    

    @property
    def Vgen_buffer(self):
        """Stored generator voltage data"""
        return self._vgen_buffer

    @property
    def Vbeam_buffer(self):
        """Stored beam induced voltage data"""
        return self._vbeam_buffer

    @property
    def Vbunch_buffer(self):
        """Stored bunch induced voltage data"""
        return numpy.moveaxis(self._vbunch_buffer, 0, -1)    

    @property
    def Nslice(self):
        """Number of slices per bunch"""
        return self._nslice

    @Nslice.setter
    def Nslice(self, nslice):
        self._nslice = nslice
        self.clear_history()

    @property
    def Nturns(self):
        """Number of turn for the wake field"""
        return self._nturns

    @Nturns.setter
    def Nturns(self, nturns):
        self._nturns = nturns
        self.clear_history()

    @property
    def TurnHistory(self):
        """Turn history of the slices center of mass"""
        return self._turnhistory

    @property
    def ResFrequency(self):
        """Resonator frequency"""
        return self.Frequency/(1-numpy.tan(self.Vgen[1])/(2*self.Qfactor))

    @property
    def Vbeam(self):
        """Beam phasor (amplitude, phase)"""
        return self._vbeam

    @property
    def Vbunch(self):
        """Bunch phasor (amplitude, phase)"""
        return self._vbunch

    @property
    def Vcav(self):
        """Cavity phasor (amplitude, phase)"""
        return self._vcav

    @property
    def Vgen(self):
        """Generator phasor (amplitude, phase)"""
        return self._vgen

    @Vgen.setter
    def Vgen(self, value):
        self._vgen = value

    @staticmethod
    def build_from_cav(cav: RFCavity, ring: Sequence,
                       qfactor: float, rshunt: float,
                       blmode: Optional[BLMode] = BLMode.PHASOR,
                       cavitymode: Optional[CavityMode] = CavityMode.ACTIVE,
                       buffersize: Optional[int] = 0, **kwargs):
        r"""Function to build the BeamLoadingElement from a cavity
        the FamName, Length, Voltage, Frequency and HarmNumber are
        taken from the cavity element

        Parameters:
            ring:            Lattice object
            qfactor:         Q factor
            rshunt:          Shunt impedance, [:math:`\Omega`]

        Keyword Arguments:
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (float):   Normalization factor
            blmode (BLMode):  method for beam loading calculation BLMode.PHASOR
                (default) uses the phasor method, BLMode.WAKE uses the wake
                function. For high Q resonator, the phasor method should be
                used
            cavitymode (CavityMode):  type of beam loaded cavity ACTIVE
                (default) for a cavity with active compensation, or
                PASSIVE to only include the beam induced voltage
            buffersize (int):  Size of the history buffer for vbeam, vgen, vbunch
                (default 0)
                
        Returns:
            bl_elem (Element): beam loading element
        """
        _CAV_ATTRIBUTES = ['Length', 'Voltage', 'Frequency']
        _EXCL_ATTRIBUTES = ['PassMethod', 'Energy', 'HarmNumber']
        cav_attrs = dict(cav.items())
        family_name = cav_attrs.pop('FamName') + '_BL'
        [cav_attrs.pop(k) for k in _EXCL_ATTRIBUTES]
        cav_args = [cav_attrs.pop(k, getattr(cav, k)) for k in
                    _CAV_ATTRIBUTES]
        if cavitymode == CavityMode.PASSIVE:
            if cav_args[1] != 0.0:
                warnings.warn(AtWarning('Setting Cavity Voltage to 0'))
            cav_args[1] = 0.0
        return BeamLoadingElement(family_name, *cav_args, ring,
                                  qfactor, rshunt, blmode=blmode,
                                  cavitymode=cavitymode,
                                  buffersize=buffersize,
                                  **cav_attrs, **kwargs)

    def __repr__(self):
        """Simplified __repr__ to avoid errors due to arguments
        not defined as attributes
        """
        attrs = dict((k, v) for (k, v) in self.items()
                     if not k.startswith('_'))
        return '{0}({1})'.format(self.__class__.__name__, attrs)



import pytest
import numpy
from at import element_track, lattice_track
from at import lattice_pass, internal_lpass
from at import element_pass, internal_epass
from at import elements
from numpy.testing import assert_equal


def test_data_checks():
    val = numpy.zeros([6, 6])
    assert elements._array(val).shape == (36,)
    assert elements._array66(val).shape == (6, 6)


def test_element_string_ordering():
    d = elements.Drift('D0', 1, attr=numpy.array(0))
    assert d.__str__() == ("Drift:\n       FamName: D0\n        Length: 1.0\n"
                           "    PassMethod: DriftPass\n          attr: 0")
    assert d.__repr__() == "Drift('D0', 1.0, attr=array(0))"


def test_element_creation_raises_exception():
    with pytest.raises(ValueError):
        elements.Element('family_name', R1='not_an_array')


def test_base_element_methods():
    e = elements.Element('family_name')
    assert e.divide([0.2, 0.5, 0.3]) == [e]
    assert id(e.copy()) != id(e)


def test_argument_checks():
    q = elements.Quadrupole('quad', 1.0, 0.5)
    # Test type
    with pytest.raises(ValueError):
        q.Length = 'a'
    # Test shape
    with pytest.raises(ValueError):
        q.T1 = [0.0, 0.0]
    # Test coherence of polynoms
    with pytest.raises(ValueError):
        q.MaxOrder = 2
    with pytest.raises(ValueError):
        q.PolynomA = [0.0]
    with pytest.raises(ValueError):
        q.PolynomB = [0.0]


def test_dipole():
    d = elements.Dipole('dipole', 1.0, 0.01)
    assert d.MaxOrder == 0
    assert len(d.PolynomA) == 2
    assert d.K == 0.0
    d = elements.Dipole('dipole', 1.0, 0.01, -0.5)
    assert d.MaxOrder == 1
    assert len(d.PolynomA) == 2
    assert d.K == -0.5
    d = elements.Dipole('dipole', 1.0, 0.01, PolynomB=[0.0, 0.1, 0.0])
    assert d.MaxOrder == 1
    assert len(d.PolynomA) == 3
    assert d.K == 0.1
    d = elements.Dipole('dipole', 1.0, 0.01, PolynomB=[0.0, 0.0, 0.005])
    assert d.MaxOrder == 2
    assert len(d.PolynomA) == 3
    assert d.K == 0.0
    d = elements.Dipole('dipole', 1.0, 0.01, PolynomB=[0.0, 0.0, 0.005],
                        MaxOrder=0)
    assert d.MaxOrder == 0
    assert len(d.PolynomA) == 3
    assert d.K == 0.0


def test_quadrupole():
    q = elements.Quadrupole('quadrupole', 1.0)
    assert q.MaxOrder == 1
    assert len(q.PolynomA) == 2
    assert q.K == 0.0
    q = elements.Quadrupole('quadrupole', 1.0, -0.5)
    assert q.MaxOrder == 1
    assert len(q.PolynomA) == 2
    assert q.K == -0.5
    q = elements.Quadrupole('quadrupole', 1.0, PolynomB=[0.0, 0.0, 0.005])
    assert q.MaxOrder == 2
    assert len(q.PolynomA) == 3
    assert q.K == 0.0
    q = elements.Quadrupole('quadrupole', 1.0, PolynomB=[0.0, 0.5, 0.005],
                            MaxOrder=1)
    assert q.MaxOrder == 1
    assert len(q.PolynomA) == 3
    assert q.K == 0.5


def test_sextupole():
    s = elements.Sextupole('sextupole', 1.0)
    assert s.MaxOrder == 2
    assert len(s.PolynomA) == 3
    assert s.H == 0.0
    s = elements.Sextupole('sextupole', 1.0, -0.5)
    assert s.MaxOrder == 2
    assert len(s.PolynomA) == 3
    assert s.H == -0.5
    s = elements.Sextupole('sextupole', 1.0, PolynomB=[0.0, 0.0, 0.005, 0.0])
    assert s.MaxOrder == 2
    assert len(s.PolynomA) == 4
    assert s.H == 0.005
    s = elements.Sextupole('sextupole', 1.0, PolynomB=[0.0, 0.0, 0.005, 0.001])
    assert s.MaxOrder == 3
    assert len(s.PolynomA) == 4
    assert s.H == 0.005
    s = elements.Sextupole('sextupole', 1.0, PolynomB=[0.0, 0.5, 0.005, 0.001],
                           MaxOrder=2)
    assert s.MaxOrder == 2
    assert len(s.PolynomA) == 4
    assert s.H == 0.005


def test_octupole():
    o = elements.Octupole('octupole', 1.0, [], [0.0, 0.0, 0.0, 0.0])
    assert o.MaxOrder == 3
    assert len(o.PolynomA) == 4


def test_thinmultipole():
    m = elements.ThinMultipole('thin', [], [0.0, 0.0, 0.0, 0.0])
    assert m.MaxOrder == 0
    assert len(m.PolynomA) == 4
    m = elements.ThinMultipole('thin', [], [0.0, 0.0, 1.0, 0.0])
    assert m.MaxOrder == 2
    assert len(m.PolynomA) == 4


def test_multipole():
    m = elements.Multipole('multi', 1.0, [], [0.0, 0.0, 0.0, 0.0])
    assert m.Length == 1.0
    assert m.MaxOrder == 0
    assert m.NumIntSteps == 10
    assert m.PassMethod == 'StrMPoleSymplectic4Pass'


def test_divide_splits_attributes_correctly():
    pre = elements.Drift('drift', 1)
    post = pre.divide([0.2, 0.5, 0.3])
    assert len(post) == 3
    assert sum([e.Length for e in post]) == pre.Length
    pre = elements.Dipole('dipole', 1, KickAngle=[0.5, -0.5], BendingAngle=0.2)
    post = pre.divide([0.2, 0.5, 0.3])
    assert len(post) == 3
    assert sum([e.Length for e in post]) == pre.Length
    assert sum([e.KickAngle[0] for e in post]) == pre.KickAngle[0]
    assert sum([e.KickAngle[1] for e in post]) == pre.KickAngle[1]
    assert sum([e.BendingAngle for e in post]) == pre.BendingAngle
    pre = elements.RFCavity('rfc', 1, voltage=187500, frequency=3.5237e+8,
                            harmonic_number=31, energy=6.e+9)
    post = pre.divide([0.2, 0.5, 0.3])
    assert len(post) == 3
    assert sum([e.Length for e in post]) == pre.Length
    assert sum([e.Voltage for e in post]) == pre.Voltage


def test_insert_into_drift():
    # Create elements
    drift = elements.Drift('drift', 1)
    monitor = elements.Monitor('bpm')
    quad = elements.Quadrupole('quad', 0.3)
    # Test None splitting behaviour
    el_list = drift.insert([(0., None), (0.3, None), (0.7, None), (1., None)])
    assert len(el_list) == 3
    numpy.testing.assert_almost_equal([e.Length for e in el_list],
                                      [0.3, 0.4, 0.3])
    # Test normal insertion
    el_list = drift.insert([(0.3, monitor), (0.7, quad)])
    assert len(el_list) == 5
    numpy.testing.assert_almost_equal([e.Length for e in el_list],
                                      [0.3, 0.0, 0.25, 0.3, 0.15])
    # Test insertion at either end produces -ve length drifts
    el_list = drift.insert([(0.0, quad), (1.0, quad)])
    assert len(el_list) == 5
    numpy.testing.assert_almost_equal([e.Length for e in el_list],
                                      [-0.15, 0.3, 0.7, 0.3, -0.15])


@pytest.mark.parametrize('func', (lattice_track, lattice_pass, internal_lpass))
def test_correct_dimensions_does_not_raise_error(rin, func):
    func([], rin, 1)
    rin = numpy.zeros((6,))
    func([], rin, 1)
    rin = numpy.array(numpy.zeros((6, 2), order='F'))
    func([], rin, 1)


@pytest.mark.parametrize("dipole_class", (elements.Dipole, elements.Bend))
@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_dipole_bend_synonym(rin, dipole_class, func):
    b = dipole_class('dipole', 1.0, 0.1, EntranceAngle=0.05, ExitAngle=0.05)
    rin[0, 0] = 1e-6
    if func == element_track:
        func(b, rin, in_place=True)
    else:
        func(b, rin)
    rin_expected = numpy.array([1e-6, 0, 0, 0, 0, 1e-7]).reshape((6, 1))
    numpy.testing.assert_almost_equal(rin, rin_expected)
    assert b.K == 0.0
    b.PolynomB[1] = 0.2
    assert b.K == 0.2
    b.K = 0.1
    assert b.PolynomB[1] == 0.1


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_marker(rin, func):
    m = elements.Marker('marker')
    assert m.Length == 0
    rin = numpy.array(numpy.random.rand(*rin.shape), order='F')
    rin_orig = numpy.array(rin, copy=True, order='F')
    if func == element_track:
        func(m, rin, in_place=True)
    else:
        func(m, rin)
    numpy.testing.assert_equal(rin, rin_orig)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_monitor(rin, func):
    mon = elements.Monitor('monitor')
    assert mon.Length == 0
    rin = numpy.array(numpy.random.rand(*rin.shape), order='F')
    rin_orig = rin.copy()
    if func == element_track:
        func(mon, rin, in_place=True)
    else:
        func(mon, rin)
    numpy.testing.assert_equal(rin, rin_orig)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_aperture_inside_limits(rin, func):
    a = elements.Aperture('aperture', [-1e-3, 1e-3, -1e-4, 1e-4])
    assert a.Length == 0
    rin[0, 0] = 1e-5
    rin[2, 0] = -1e-5
    rin_orig = rin.copy()
    if func == element_track:
        func(a, rin, in_place=True)
    else:
        func(a, rin)
    numpy.testing.assert_equal(rin, rin_orig)


@pytest.mark.parametrize('func', (lattice_track, lattice_pass, internal_lpass))
def test_aperture_outside_limits(rin, func):
    a = elements.Aperture('aperture', [-1e-3, 1e-3, -1e-4, 1e-4])
    assert a.Length == 0
    lattice = [a]
    rin[0, 0] = 1e-2
    rin[2, 0] = -1e-2
    if func == lattice_track:
        func(lattice, rin, in_place=True)
    else:
        func(lattice, rin)
    assert numpy.isnan(rin[0, 0])
    assert rin[2, 0] == 0.0  # Only the 1st coordinate is nan, the rest is zero


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_drift_offset(rin, func):
    d = elements.Drift('drift', 1)
    rin[0, 0] = 1e-6
    rin[2, 0] = 2e-6
    rin_orig = rin.copy()
    if func == element_track:
        func(d, rin, in_place=True)
    else:
        func(d, rin)
    numpy.testing.assert_equal(rin, rin_orig)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_drift_divergence(rin, func):
    d = elements.Drift('drift', 1.0)
    assert d.Length == 1
    rin[1, 0] = 1e-6
    rin[3, 0] = -2e-6
    if func == element_track:
        func(d, rin, in_place=True)
    else:
        func(d, rin)
    # results from Matlab
    rin_expected = numpy.array([1e-6, 1e-6, -2e-6, -2e-6, 0,
                                2.5e-12]).reshape(6, 1)
    numpy.testing.assert_equal(rin, rin_expected)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_drift_two_particles(rin, func):
    d = elements.Drift('drift', 1.0)
    assert d.Length == 1
    two_rin = numpy.array(numpy.concatenate((rin, rin), axis=1), order='F')
    # particle one is offset
    two_rin[0, 0] = 1e-6
    two_rin[2, 0] = 2e-6
    # particle two has divergence
    two_rin[1, 1] = 1e-6
    two_rin[3, 1] = -2e-6
    two_rin_orig = two_rin.copy()
    if func == element_track:
        func(d, two_rin, in_place=True)
    else:
        func(d, two_rin)
    # results from Matlab
    p1_expected = numpy.array(two_rin_orig[:, 0]).reshape(6, 1)
    p2_expected = numpy.array([1e-6, 1e-6, -2e-6, -2e-6, 0,
                               2.5e-12]).reshape(6, 1)
    two_rin_expected = numpy.concatenate((p1_expected, p2_expected), axis=1)
    numpy.testing.assert_equal(two_rin, two_rin_expected)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_quad(rin, func):
    q = elements.Quadrupole('quad', 0.4, k=1)
    rin[0, 0] = 1e-6
    if func == element_track:
        func(q, rin, in_place=True)
    else:
        func(q, rin)
    expected = numpy.array([0.9210610203854122, -0.3894182419439, 0,
                            0, 0, 0.0000000103303797478]).reshape(6, 1) * 1e-6
    numpy.testing.assert_allclose(rin, expected)
    assert q.K == 1
    q.PolynomB[1] = 0.2
    assert q.K == 0.2
    q.K = 0.1
    assert q.PolynomB[1] == 0.1


@pytest.mark.parametrize('func', (lattice_track, lattice_pass, internal_lpass))
def test_rfcavity(rin, func):
    rf = elements.RFCavity('rfcavity', 0.0, 187500, 3.5237e+8, 31, 6.e+9)
    lattice = [rf, rf, rf, rf]
    rin[4, 0] = 1e-6
    rin[5, 0] = 1e-6
    if func == lattice_track:
        func(lattice, rin, in_place=True)
    else:
        func(lattice, rin)
    expected = numpy.array([0., 0., 0., 0., 9.990769e-7, 1.e-6]).reshape(6, 1)
    numpy.testing.assert_allclose(rin, expected, atol=1e-12)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
@pytest.mark.parametrize("n", (0, 1, 2, 3, 4, 5))
def test_m66(rin, n, func):
    m = numpy.random.rand(6, 6)
    m66 = elements.M66('m66', m)
    assert m66.Length == 0
    rin[n, 0] = 1e-6
    if func == element_track:
        func(m66, rin, in_place=True)
    else:
        func(m66, rin)
    expected = numpy.array([m[0, n], m[1, n], m[2, n], m[3, n], m[4, n],
                            m[5, n]]).reshape(6, 1) * 1e-6
    numpy.testing.assert_equal(rin, expected)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_corrector(rin, func):
    c = elements.Corrector('corrector', 0.0, numpy.array([0.9, 0.5],
                                                         dtype=numpy.float64))
    assert c.Length == 0
    rin[0, 0] = 1e-6
    rin_orig = rin.copy()
    rin_orig[1] = 0.9
    rin_orig[3] = 0.5
    if func == element_track:
        func(c, rin, in_place=True)
    else:
        func(c, rin)
    numpy.testing.assert_equal(rin, rin_orig)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_wiggler(rin, func):
    period = 0.05
    periods = 23
    bmax = 1
    by = numpy.array([1, 1, 0, 1, 1, 0], dtype=numpy.float64)
    c = elements.Wiggler('wiggler', period * periods, period, bmax, 3e9, By=by)
    assert abs(c.Length - 1.15) < 1e-10
    # Expected value from Matlab AT.
    expected = numpy.array(rin, copy=True)
    expected[5] = 0.000000181809691064259
    if func == element_track:
        func(c, rin, in_place=True)
    else:
        func(c, rin)
    numpy.testing.assert_allclose(rin, expected, atol=1e-12)


def test_exit_entrance():
    q = elements.Quadrupole('quad', 0.4, k=1)
    for kin, kout in zip(q._entrance_fields, q._exit_fields):
        assert_equal(kin.replace('Entrance', ''). replace('1', ''),
                     kout.replace('Exit', '').replace('2', ''))



from at import physics, get_s_pos, uint32_refpts
from scipy.constants import speed_of_light
import matlab
import numpy as np
from numpy.testing import assert_allclose as assert_close
import pytest


def _py_data(ml_data):
    """Convert a Matlab vector to a numpy vector"""
    return np.squeeze(np.asarray(ml_data))


def _ml_refs(refpts, nelems):
    """Convert refpoints to Matlab"""
    uintrefs = uint32_refpts(refpts, nelems)
    return matlab.double([float(ref + 1) for ref in uintrefs])


def _compare_physdata(py_data, ml_data, fields, atol=12, rtol=1.0e-7):
    for ml_key, py_key in fields:
        ml_val = _py_data(ml_data[ml_key])
        if py_key == "closed_orbit":
            py_val = np.squeeze(py_data[py_key][:, :4])
        else:
            py_val = np.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


@pytest.mark.parametrize("dp", (-0.01, 0.0, 0.01))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_linear_analysis(engine, request, lattices, dp):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    fields = [
        ("SPos", "s_pos"),
        ("ClosedOrbit", "closed_orbit"),
        ("Dispersion", "dispersion"),
        ("alpha", "alpha"),
        ("beta", "beta"),
        ("mu", "mu"),
        ("M44", "m44"),
        ("A", "A"),
        ("B", "B"),
        ("C", "C"),
        ("gamma", "gamma"),
    ]
    refpts = range(nelems + 1)
    py_data0, py_tune, py_chrom, py_data = physics.linopt(py_lattice, dp, refpts, True)
    # Matlab call
    ml_data, ml_tune, ml_chrom = engine.pyproxy(
        "atlinopt", ml_lattice, dp, _ml_refs(refpts, nelems), nargout=3
    )
    ml_data0 = engine.pyproxy(
        "atlinopt", ml_lattice, dp, _ml_refs(nelems, nelems), nargout=3
    )[0]
    # Comparison
    assert_close(py_tune, _py_data(ml_tune), atol=1.0e-8, rtol=0)
    assert_close(py_chrom, _py_data(ml_chrom), atol=3.0e-3, rtol=0)
    _compare_physdata(np.expand_dims(py_data0, 0), ml_data0, fields, rtol=1e-8)
    _compare_physdata(py_data, ml_data, fields, rtol=1e-8)


@pytest.mark.parametrize("lattices", ["hmba"])
def test_ohmi_envelope(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    fields = [
        ("beam66", "r66"),
        ("beam44", "r44"),
        ("emit66", "emitXYZ"),
        ("emit44", "emitXY"),
    ]
    nelems = len(py_lattice)
    refpts = range(nelems + 1)

    # Python call
    py_lattice = py_lattice.radiation_on(copy=True)
    py_emit0, py_beamdata, py_emit = py_lattice.ohmi_envelope(refpts)
    # Matlab call
    ml_emit = engine.pyproxy("atx", ml_lattice, 0.0, _ml_refs(refpts, nelems))
    ml_emit0, ml_params = engine.pyproxy(
        "atx", ml_lattice, 0.0, _ml_refs(0, nelems), nargout=2
    )
    revolution_period = get_s_pos(py_lattice, nelems)[0] / speed_of_light
    damping_times = revolution_period / py_beamdata.damping_rates
    # Comparison
    assert_close(
        damping_times, _py_data(ml_params["dampingtime"]), atol=1.0e-6, rtol=1.0e-3
    )
    assert_close(
        py_beamdata.mode_emittances,
        _py_data(ml_emit0["modemit"]),
        atol=1.0e-20,
        rtol=2.0e-3,
    )
    _compare_physdata(np.expand_dims(py_emit0, 0), ml_emit0, fields)
    _compare_physdata(py_emit, ml_emit, fields)


@pytest.mark.parametrize("lattices", ["hmba_rad"])
def test_quantdiff(engine, request, lattices):
    py_lattice, ml_lattice, radindex = request.getfixturevalue(lattices)
    # Python call
    dmat = physics.radiation.quantdiffmat(py_lattice)
    # Matlab call
    dmat_ml = engine.pyproxy("quantumDiff", ml_lattice, radindex)
    # Comparison
    assert_close(dmat, dmat_ml, rtol=1.0e-8, atol=1.0e-18)


@pytest.mark.parametrize("lattices", ["hmba"])
def test_fastring(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    # Python call
    ring, ringrad = physics.fastring.fast_ring(py_lattice)
    # Matlab call
    ring_ml, ringrad_ml = engine.pyproxy("atfastring", ml_lattice, nargout=2)
    # Comparison
    for r, rml in zip([ring, ringrad], [ring_ml, ringrad_ml]):
        assert_close(r[0].Frequency, rml[1]["Frequency"], rtol=1.0e-20)
        assert_close(r[0].Voltage, rml[1]["Voltage"], rtol=1.0e-20)
        assert_close(r[1].I2, rml[2]["I2"], rtol=1.0e-20)
        assert_close(r[1].Length, rml[2]["Length"], rtol=1.0e-20)
        assert_close(r[1].M66, rml[2]["M66"], atol=1.0e-7)
        assert_close(r[1].T1, np.squeeze(rml[2]["T1"]), rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[1].T2, np.squeeze(rml[2]["T2"]), rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[2].A1, rml[-1]["A1"], rtol=0.02)
        assert_close(r[2].A2, rml[-1]["A2"], rtol=0.02)
        assert_close(r[2].A3, rml[-1]["A3"], rtol=0.02)
        assert_close(r[2].Alphax, rml[-1]["Alphax"], rtol=1.0e-5)
        assert_close(r[2].Alphay, rml[-1]["Alphay"], rtol=1.0e-6)
        assert_close(r[2].Betax, rml[-1]["Betax"], rtol=1.0e-10)
        assert_close(r[2].Betay, rml[-1]["Betay"], rtol=1.0e-10)
        assert_close(r[2].Qpx, rml[-1]["Qpx"], rtol=1.0e-5)
        assert_close(r[2].Qpy, rml[-1]["Qpy"], rtol=1.0e-5)
        assert_close(r[2].T1, np.squeeze(rml[-1]["T1"]), rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[2].T2, np.squeeze(rml[-1]["T2"]), rtol=1.0e-8, atol=1.0e-11)
        if len(r) >= 4:
            assert_close(r[3].Lmatp, rml[-2]["Lmatp"], rtol=0.02, atol=1.0e-10)


@pytest.mark.parametrize("dp", (0.00, 0.01, -0.01))
@pytest.mark.parametrize("lattices", ["hmba"])
def test_parameters(engine, request, lattices, dp):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)

    # Test perimeter
    py_length = py_lattice.get_s_pos(len(py_lattice))
    ml_length = engine.findspos(ml_lattice, len(ml_lattice) + 1)
    assert_close(py_length, ml_length, rtol=1.0e-8)

    # test energy loss
    ml_energy, ml_periods, ml_voltage, ml_harms, ml_eloss = engine.pyproxy(
        "atenergy", ml_lattice, nargout=5
    )
    assert_close(py_lattice.rf_voltage, ml_voltage, rtol=1.0e-8)
    assert_close(py_lattice.energy_loss, ml_eloss, rtol=1.0e-6)
    assert py_lattice.energy == ml_energy
    assert py_lattice.periodicity == int(ml_periods)
    assert py_lattice.harmonic_number == int(ml_harms)

    # test momentum compaction factor
    py_mcf = py_lattice.get_mcf(dp)
    ml_mcf = engine.mcf(ml_lattice, dp)
    assert_close(py_mcf, ml_mcf, rtol=1.0e-8)



import glob
import os
from os.path import basename, exists, join, splitext
import sys
from setuptools import setup, Extension

# Numpy build dependency defined in pyproject.toml.
import numpy


def select_omp():
    if exists('/usr/local/include/omp.h'):              # Homebrew
        return '-I/usr/local/include', '/usr/local/lib'
    elif exists('/opt/local/include/libomp/omp.h'):     # MacPorts
        return '-I/opt/local/include/libomp', '/opt/local/lib/libomp'
    else:
        raise FileNotFoundError('\n'.join((
            '',
            'libomp.dylib must be installed with your package manager:',
            '',
            'Use "$ brew install libomp"',
            'Or  "$ sudo port install libomp"',
            ''
        )))


print("** Entering setup.py:", str(sys.argv))
print("** MPI:", os.environ.get('MPI', None))
print("** OPENMP:", os.environ.get('OPENMP', None))
macros = [('PYAT', None)]
with_openMP = False

cflags = ['-std=c99']
cppflags = []

mpi = eval(os.environ.get('MPI', 'None'))
if not mpi or (len(sys.argv) >= 2 and
               any(sys.argv[1] == arg for arg in ('egg_info', 'sdist'))):
    mpi_macros = []
    mpi_includes = []
else:
    mpi_macros = [('MPI', None)]
    os.environ["CC"] = 'mpicc'
    os.environ["CXX"] = 'mpicxx'
    try:
        import mpi4py
    except ImportError:
        print('\npyAT with MPI requires mpi4py. '
              'Please install mpi4py: "pip install mpi4py"\n')
        sys.exit()
    mpi_includes = mpi4py.get_include()


omp = eval(os.environ.get('OPENMP', 'None'))
if not omp:
    omp_cflags = []
    omp_lflags = []
    omp_macros = []
else:
    # Get the location of an alternate OpenMP library
    # Example: OMP_MATLAB=$MATLABROOT/sys/os/glnx64
    omp_path = os.environ.get('OMP_MATLAB', None)
    # Get the threshold on the number of particles
    omp_threshold = int(os.environ.get('OMP_PARTICLE_THRESHOLD', 10))
    omp_macros = [('OMP_PARTICLE_THRESHOLD', omp_threshold)]
    if sys.platform.startswith('win'):
        omp_cflags = ['/openmp']
        omp_lflags = []
    elif sys.platform.startswith('darwin'):
        omp_inc, omp_lib = select_omp()
        omp_cflags = ['-Xpreprocessor', '-fopenmp', omp_inc]
        if omp_path is None:
            omp_lflags = ['-L' + omp_lib, '-lomp']
        else:
            omp_lflags = ['-L' + omp_path, '-Wl,-rpath,' + omp_path, '-liomp5']
    else:
        omp_cflags = ['-fopenmp']
        if omp_path is None:
            omp_lflags = ['-lgomp']
        else:
            omp_lflags = ['-L' + omp_path, '-Wl,-rpath,' + omp_path, '-liomp5']

if not sys.platform.startswith('win32'):
    cflags += ['-Wno-unused-function']

# It is easier to copy the integrator files into a directory inside pyat
# for packaging. However, we cannot always rely on this directory being
# inside a clone of at and having the integrator sources available, and
# this file is executed each time any setup.py command is run.
# It appears that only copying the files when they are available is
# sufficient.
integrator_src_orig = 'atintegrators'
diffmatrix_orig = join('atmat', 'atphysics', 'Radiation')

c_pass_methods = glob.glob(join(integrator_src_orig, '*Pass.c'))
cpp_pass_methods = glob.glob(join(integrator_src_orig, '*Pass.cc'))
diffmatrix_source = join(diffmatrix_orig, 'findmpoleraddiffmatrix.c')
at_source = join('pyat', 'at.c')


def c_integrator_ext(pass_method):
    name, _ = splitext(basename(pass_method))
    name = ".".join(('at', 'integrators', name))
    return Extension(
        name=name,
        sources=[pass_method],
        include_dirs=[numpy.get_include(), mpi_includes, integrator_src_orig],
        define_macros=macros + omp_macros + mpi_macros,
        extra_compile_args=cflags + omp_cflags,
        extra_link_args=omp_lflags
    )


def cpp_integrator_ext(pass_method):
    name, _ = splitext(basename(pass_method))
    name = ".".join(('at', 'integrators', name))
    return Extension(
        name=name,
        sources=[pass_method],
        include_dirs=[numpy.get_include(), mpi_includes, integrator_src_orig],
        define_macros=macros + omp_macros + mpi_macros,
        extra_compile_args=cppflags + omp_cflags,
        extra_link_args=omp_lflags
    )


at = Extension(
    'at.tracking.atpass',
    sources=[at_source],
    define_macros=macros + omp_macros + mpi_macros,
    include_dirs=[numpy.get_include(), integrator_src_orig],
    extra_compile_args=cflags + omp_cflags,
    extra_link_args=omp_lflags
)

cconfig = Extension(
    'at.cconfig',
    sources=[join('pyat', 'at', 'cconfig.c')],
    define_macros=macros + omp_macros + mpi_macros,
    extra_compile_args=cflags + omp_cflags,
)

diffmatrix = Extension(
    name='at.physics.diffmatrix',
    sources=[diffmatrix_source],
    include_dirs=[numpy.get_include(), integrator_src_orig],
    define_macros=macros,
    extra_compile_args=cflags
)

setup(
    ext_modules=[at, cconfig, diffmatrix] +
                [c_integrator_ext(pm) for pm in c_pass_methods] +
                [cpp_integrator_ext(pm) for pm in cpp_pass_methods],
)

print("** Leaving setup.py")



"""
Frequency analysis (FMAP) using .harmonic_analysis lib
and pyat parallel tracking (patpass)
"""

# orblancog
# generates the frequency and diffusion map for a given ring
# 2023jan16 tracking is parallel (patpass), frequency analysis is serial
# 2022jun07 serial version

from at.tracking import patpass
from at.physics import find_orbit
import numpy
from warnings import warn
from at.lattice import AtWarning


# Jaime Coello de Portugal (JCdP) frequency analysis implementation
from .harmonic_analysis import get_tunes_harmonic

__all__ = ['fmap_parallel_track']


def fmap_parallel_track(ring,
                        coords=[-10, 10, -10, 10],
                        steps=[100, 100],
                        scale='linear',
                        turns=512,
                        orbit=None,
                        add_offset6D=numpy.zeros(6),
                        verbose=False,
                        lossmap=False,
                        **kwargs
                        ):
    r"""Computes frequency maps

    This function calculates the norm of the transverse tune variation per turn
    for a particle tracked along a ring with a set of offsets in the initial
    coordinates.

    It returns a numpy array containing 7 columns:

    | ``[xoffset, yoffset, nux, nuy, dnux, dnuy,``
    | ``log10(sqrt(dnux*dnux + dnuy*dnuy)/tns )]``

    for every tracked particle that survives 2*turns.
    Particles lost before 2*turns are ignored.
    The log10 tune variation is limited to the interval from -10 to -2;
    values above or below are set to the closer limit.

    The transverse offsets are given inside a rectangular coordinate window

    ``coords=[xmin, xmax, ymin, *ymax]``

    in millimeters, where the window is divided in n steps=[xsteps,ysteps].
    For each yoffset, particle tracking over the whole set of xoffsets is done
    in parallel.

    The frequency analysis uses a library that is not parallelized.

    The closed orbit is calculated and added to the
    initial particle offset of every particle. Otherwise, one could set
    ``orbit = numpy.zeros(6)`` to avoid it.
    Additionally, a numpy array (*add_offset6D*) with 6 values could be used to
    arbitrarily offset the initial coordinates of every particle.

    A dictionary with particle losses is saved for every vertical offset.
    See the :py:func:`.patpass` documentation.

    Parameters:
        ring:     a valid pyat ring
        coords:   default [-10,10,-10,10] in mm
        steps:    (xsteps, ysteps): number of steps in each plane
        scale:    default 'linear'; or 'non-linear'
        turns:    default 512
        orbit:    If :py:obj:`None`, the closed orbit is computed and added to
          the coordinates
        add_offset6D: default numpy.zeros((6,1))
        verbose:  prints additional info
        lossmap:  default false
    Optional:
        pool_size:number of processes. See :py:func:`.patpass`

    Returns:
        xy_nuxy_lognudiff_array: numpy array with columns
         `[xcoor, ycoor, nux, ny, dnux, dnuy,
         log10(sqrt(sum(dnu**2)/turns)) ]`
        loss_map_array : experimental format.
          if *loss_map* is :py:obj:`True`, it returns the losses dictionary
          provided by :py:func:`.patpass` per every vertical offset.
          if *loss_map* is :py:obj:`False`, it returns a one-element list.

    .. warning::

       points with NaN tracking results or non-defined frequency in x,y
       are ignored.

    .. warning:: loss map format is experimental
    """

    # https://github.com/atcollab/at/pull/608
    kwargs.setdefault('pool_size', None)

    # https://github.com/atcollab/at/pull/608
    if 'ncpu' in kwargs:
        warn(AtWarning('ncpu argument is deprecated; use pool_size instead'))
        kwargs['pool_size'] = kwargs.pop('ncpu')

    if orbit is None:
        # closed orbit values are not returned. It seems not necessary here
        orbit, _ = find_orbit(ring)
        print(f'Closed orbit:\t{orbit}')

    # verboseprint to check flag only once
    verboseprint = print if verbose else lambda *a, **k: None

    # tns is the variable used in the frequency analysis
    # turns is the input variable from user
    # nturns is twice tns in order to get the tune in the first and second
    #     part of the tracking
    tns = turns
    nturns = 2*tns

    # scale the inumpyut coordinates
    xscale = 1e-3
    yscale = 1e-3

    # returned array
    xy_nuxy_lognudiff_array = numpy.empty([])
    loss_map_array = numpy.empty([])

    # verify steps
    if numpy.count_nonzero(steps) != 2:
        raise ValueError('steps can not be zero')

    xmin = numpy.minimum(coords[0], coords[1])
    xmax = numpy.maximum(coords[0], coords[1])
    ymin = numpy.minimum(coords[2], coords[3])
    ymax = numpy.maximum(coords[2], coords[3])
    xsteps = steps[0]
    ysteps = steps[1]

    # define rectangle and x,y step size
    if scale == "nonlinear":
        verboseprint('non linear steps')
        xinterval = 1.0*(xmax - xmin)
        yinterval = 1.0*(ymax - ymin)
        xauxsteps = numpy.arange(-0.5, 0.0 + 1e-9, 1./xsteps)
        yauxsteps = numpy.arange(-0.5, 0.0 + 1e-9, 1./ysteps)
        xnrsteps = 10**xauxsteps
        xnrsteps = xnrsteps / sum(xnrsteps)
        ynrsteps = 10**yauxsteps
        ynrsteps = ynrsteps / sum(ynrsteps)
        xscalesteps = numpy.concatenate((xnrsteps,
                                        numpy.flip(xnrsteps)), axis=None)
        yscalesteps = numpy.concatenate((ynrsteps,
                                        numpy.flip(ynrsteps)), axis=None)
        ixarray = xmin + 0.5 * xinterval * numpy.cumsum(xscalesteps)
        iyarray = ymin + 0.5 * yinterval * numpy.cumsum(yscalesteps)
    else:  # scale == 'linear', ignore any other
        verboseprint('linear steps')
        # get the intervals
        xstep = 1.0*(xmax - xmin)/xsteps
        ystep = 1.0*(ymax - ymin)/ysteps
        ixarray = numpy.arange(xmin, xmax+1e-6, xstep)
        iyarray = numpy.arange(ymin, ymax+1e-6, ystep)
    lenixarray = len(ixarray)
    leniyarray = len(iyarray)

    print("Start tracking and frequency analysis")

    # tracking in parallel multiple x coordinates with the same y coordinate
    for iy, iy_index in zip(iyarray, range(leniyarray)):
        print(f'Tracked particles {abs(-100.0*iy_index/leniyarray):.1f} %')
        verboseprint("y =", iy)
        z0 = numpy.zeros((lenixarray, 6))  # transposed, and C-aligned
        z0 = z0 + add_offset6D + orbit
        # add 1 nm to tracking to avoid zeros in array for the ideal lattice
        z0[:, 0] = z0[:, 0] + xscale*ixarray + 1e-9
        z0[:, 2] = z0[:, 2] + yscale*iy + 1e-9

        verboseprint("tracking ...")
        # z0.T is Fortran-aligned
        if lossmap:
            # patpass output changes when losses flag is true
            zOUT, dictloss = \
                patpass(ring, z0.T, nturns, losses=lossmap, **kwargs)
            loss_map_array = numpy.append(loss_map_array, dictloss)
        else:
            zOUT = patpass(ring, z0.T, nturns, **kwargs)

        # start of serial frequency analysis
        for ix_index in range(lenixarray):  # cycle over the track results
            # check if nan in arrays
            array_sum = numpy.sum(zOUT[:, ix_index, 0])
            array_has_nan = numpy.isnan(array_sum)
            if array_has_nan:
                verboseprint("array has nan")
                continue

            # get one valid particle
            z1 = zOUT[:, ix_index, 0]

            # remove mean values
            # get the first turn in x
            xfirst = z1[0, 0:tns]
            xfirst = xfirst - numpy.mean(xfirst)
            # pxfirst = z1[1, 0:tns]
            # pxfirst = pxfirst - numpy.mean(pxfirst)
            # xfirstpart = xfirst + 1j*pxfirst
            # get the last turns in x
            xlast = z1[0, tns:2*tns]
            xlast = xlast - numpy.mean(xlast)
            # pxlast = z1[1, tns:2*tns]
            # pxlast = pxlast - numpy.mean(pxlast)
            # xlastpart = xlast + 1j*pxlast

            # get the first turn in y
            yfirst = z1[2, 0:tns]
            yfirst = yfirst - numpy.mean(yfirst)
            # pyfirst = z1[3, 0:tns]
            # pyfirst = pyfirst - numpy.mean(pyfirst)
            # yfirstpart = yfirst + 1j*pyfirst
            # get the last turns in y
            ylast = z1[2, tns:2*tns]
            ylast = ylast - numpy.mean(ylast)
            # pylast = z1[3, tns:2*tns]
            # pylast = pylast - numpy.mean(pylast)
            # ylastpart = ylast + 1j*pylast

            # calc frequency from array,
            # jump the cycle is no frequency is found
            xfreqfirst = get_tunes_harmonic(xfirst)
            if len(xfreqfirst) == 0:
                verboseprint("  No frequency")
                continue
            # xfreqlast  = PyNAFF.naff(xlastpart,tns,1,0,False)
            xfreqlast = get_tunes_harmonic(xlast)
            if len(xfreqlast) == 0:
                verboseprint("  No frequency")
                continue
            # yfreqfirst = PyNAFF.naff(yfirstpart,tns,1,0,False)
            yfreqfirst = get_tunes_harmonic(yfirst)
            if len(yfreqfirst) == 0:
                verboseprint("  No frequency")
                continue
            # yfreqlast  = PyNAFF.naff(ylastpart,tns,1,0,False)
            yfreqlast = get_tunes_harmonic(ylast)
            if len(yfreqlast) == 0:
                verboseprint("  No frequency")
                continue
            verboseprint("NAFF results, (x,y)=",
                         ixarray[ix_index],
                         iy,
                         "\nH freq. first part =\t", xfreqfirst[0],
                         "\nH freq. last part =\t", xfreqlast[0],
                         "\nV freq. first part =\t", yfreqfirst[0],
                         "\nV freq. last part =\t", yfreqlast[0])

            # metric
            xdiff = xfreqlast[0] - xfreqfirst[0]
            ydiff = yfreqlast[0] - yfreqfirst[0]
            nudiff = 0.5*numpy.log10((xdiff*xdiff + ydiff*ydiff)/tns)
            # min max diff
            if nudiff > -2:
                nudiff = -2
            if nudiff < -10:
                nudiff = -10
            # save diff
            xy_nuxy_lognudiff_array = numpy.append(xy_nuxy_lognudiff_array,
                                                   [ixarray[ix_index],
                                                    iy,
                                                    xfreqfirst[0],
                                                    yfreqfirst[0],
                                                    xdiff,
                                                    ydiff,
                                                    nudiff])

    # first element is garbage
    xy_nuxy_lognudiff_array = numpy.delete(xy_nuxy_lognudiff_array, 0)
    # reshape for plots and output files
    xy_nuxy_lognudiff_array = xy_nuxy_lognudiff_array.reshape(-1, 7)

    return xy_nuxy_lognudiff_array, loss_map_array
# the end



"""
A special file that contains test fixtures for the other test files to use.
"""
import sys
if sys.version_info.minor < 9:
    from importlib_resources import files, as_file
else:
    from importlib.resources import files, as_file
import platform
import numpy
import pytest
import at
import machine_data
from at import elements


# noinspection PyUnusedLocal
def pytest_report_header(config):
    try:
        a = platform.uname()
        sysinfo = "system: {} {}".format(a[0], a[2])
    except AttributeError:
        sysinfo = ''
    numpyinfo = "numpy version: {}".format(numpy.__version__)
    return [sysinfo, numpyinfo]


@pytest.fixture
def rin():
    rin = numpy.array(numpy.zeros((6, 1)), order='F')
    return rin


@pytest.fixture(scope='session')
def simple_ring():
    ring = [elements.Drift('D1', 1, R1=numpy.eye(6), R2=numpy.eye(6)),
            elements.Marker('M1', attr='a_value'), elements.M66('M66'),
            elements.Drift('D2', 1, T1=numpy.zeros(6), T2=numpy.zeros(6)),
            elements.Drift('D3', 1, R1=numpy.eye(6), R2=numpy.eye(6)),
            elements.Drift('D4', 1, T1=numpy.zeros(6), T2=numpy.zeros(6))]
    return ring


@pytest.fixture(scope='session')
def simple_lattice(simple_ring):
    return at.Lattice(simple_ring, name='lat', energy=5, periodicity=1)


@pytest.fixture(scope='session')
def dba_lattice():
    with as_file(files(machine_data) / 'dba.mat') as path:
        ring = at.load_lattice(path)
    return ring


@pytest.fixture(scope='session')
def hmba_lattice():
    with as_file(files(machine_data) / 'hmba.mat') as path:
        ring = at.load_lattice(path)
    return ring



"""Load a lattice from a Tracy file (.lat).

This is not complete but can parse the example files that I have.
This parser is quite similar to the Elegant parser in elegant.py.

"""

import logging as log
from os.path import abspath
import re
import numpy
from at.lattice.elements import (
    Drift,
    Dipole,
    Quadrupole,
    Sextupole,
    Multipole,
    Corrector,
    Marker,
    RFCavity,
)
from at.lattice import Lattice
from at.load import register_format, utils

__all__ = ["load_tracy"]


def create_drift(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    return Drift(name, length, **params)


# noinspection PyUnusedLocal
def create_marker(name, params, variables):
    return Marker(name, **params)


def create_quad(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    params["NumIntSteps"] = parse_float(params.pop("n", 10), variables)
    params["k"] = parse_float(params.pop("k"), variables)
    params["PassMethod"] = "StrMPoleSymplectic4Pass"
    return Quadrupole(name, length, **params)


def create_sext(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    params["NumIntSteps"] = parse_float(params.pop("n", 10), variables)
    k2 = parse_float(params.pop("k", 0), variables)
    return Sextupole(name, length, k2, **params)


def create_dipole(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    params["NumIntSteps"] = parse_float(params.pop("n", 10), variables)
    params["PassMethod"] = "BndMPoleSymplectic4Pass"
    params["BendingAngle"] = (parse_float(params.pop("t"), variables) / 180) * numpy.pi
    params["EntranceAngle"] = (
        parse_float(params.pop("t1"), variables) / 180
    ) * numpy.pi
    params["ExitAngle"] = (parse_float(params.pop("t2"), variables) / 180) * numpy.pi
    # Tracy is encoding gap plus fringe int in the 'gap' field.
    # Since BndMPoleSymplectic4Pass only uses the product of FringeInt
    # and gap we can substitute the following.
    if "gap" in params:
        params["FullGap"] = parse_float(params.pop("gap"), variables) / 2
        params["FringeInt1"] = 1
        params["FringeInt2"] = 1
    k = parse_float(params.pop("k", 0), variables)
    if "PolynomB" in params:
        params["PolynomB"][1] = k
    else:
        params["PolynomB"] = [0, k, 0, 0]
    return Dipole(name, length, **params)


def create_corrector(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    # Is there a property for this?
    kick_angle = [0, 0]
    return Corrector(name, length, kick_angle, **params)


def create_multipole(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    params["NumIntSteps"] = parse_float(params.pop("n", 10), variables)
    poly_a = [0, 0, 0, 0]
    poly_b = [0, 0, 0, 0]
    return Multipole(name, length, poly_a, poly_b, **params)


def create_cavity(name, params, variables):
    length = parse_float(params.pop("l", 0), variables)
    voltage = parse_float(params.pop("voltage"), variables)
    frequency = parse_float(params.pop("frequency"), variables)
    params["Phi"] = parse_float(params.pop("phi", 0), variables)
    harmonic_number = variables["harmonic_number"]
    energy = variables["energy"]
    return RFCavity(name, length, voltage, frequency, harmonic_number, energy, **params)


ELEMENT_MAP = {
    "drift": create_drift,
    "bending": create_dipole,
    "quadrupole": create_quad,
    "sextupole": create_sext,
    "multipole": create_multipole,
    "corrector": create_corrector,
    "marker": create_marker,
    "map": create_marker,
    "beampositionmonitor": create_marker,
    "cavity": create_cavity,
}


def tokenise_expression(expression):
    tokens = []
    current_token = ""
    for char in expression:
        if char.isspace():  # complete the current token
            if current_token:
                tokens.append(current_token)
                current_token = ""
            continue
        if char in "+-" and current_token and current_token[-1] in "de":
            # special case for 1e-4, 1.23e+3, 2.2d6:
            # + or - are part of the current token, not their own token.
            try:
                # If d or e refers to a number, it must be possible
                # to convert the preceding characters into a float.
                float(current_token[:-1])
                current_token += char
                continue
            except ValueError:
                pass
        if char in "+-*/()":
            # standalone tokens: complete the current token and add this
            if current_token:
                tokens.append(current_token)
                current_token = ""
            tokens.append(char)
        else:
            current_token += char

    if current_token:
        tokens.append(current_token)
    return tokens


def parse_float(expression, variables):
    """Evaluate the provided arithmetic expression substituting variables."""
    log.debug("parse_float {}".format(expression))
    try:
        return float(expression)
    except ValueError:
        raw_tokens = tokenise_expression(expression)
        tokens = []
        for token in raw_tokens:
            if token in variables:
                tokens.append(variables[token])
            else:
                # Handle unusual format 2.2d6 which means 2.2e+6.
                dformat_matches = re.match("(-?\\d+.?\\d*)d(-?\\d+)", token)
                if dformat_matches:
                    mantissa, exponent = dformat_matches.groups()
                    tokens.append("{}e{}".format(mantissa, exponent))
                else:
                    tokens.append(token)

        log.debug(tokens)

        def evaluate(tokens):
            if len(tokens) == 1:
                return float(tokens[0])
            if len(tokens) == 2:
                if tokens[0] == "+":
                    return float(tokens[1])
                elif tokens[0] == "-":
                    return -float(tokens[1])

            # Remove superfluous outer parentheses.
            if tokens[0] == "(" and tokens[-1] == ")":
                return evaluate(tokens[1:-1])
            # First evaluate contents of parentheses.
            try:
                b1 = tokens.index("(")
                b2 = len(tokens) - 1 - tokens[::-1].index(")")
                return evaluate(
                    tokens[:b1] + [evaluate(tokens[b1 + 1 : b2])] + tokens[b2 + 1 :]
                )
            except ValueError:
                # No open parentheses found.
                pass
            # Evaluate / and * from left to right.
            for i, token in enumerate(tokens[:-1]):
                if token == "/":
                    return evaluate(
                        tokens[: i - 1]
                        + [float(tokens[i - 1]) / float(tokens[i + 1])]
                        + tokens[i + 2 :]
                    )
                if token == "*":
                    return evaluate(
                        tokens[: i - 1]
                        + [float(tokens[i - 1]) * float(tokens[i + 1])]
                        + tokens[i + 2 :]
                    )
            # Evaluate + and - from left to right.
            for i, token in enumerate(tokens[:-1]):
                if token == "+":
                    return evaluate(
                        tokens[: i - 1]
                        + [float(tokens[i - 1]) + float(tokens[i + 1])]
                        + tokens[i + 2 :]
                    )
                if token == "-":
                    return evaluate(
                        tokens[: i - 1]
                        + [float(tokens[i - 1]) - float(tokens[i + 1])]
                        + tokens[i + 2 :]
                    )

        return evaluate(tokens)


def parse_lines(contents):
    """Return individual lines.

    Remove comments and whitespace, convert to lowercase, and split on
    semicolons.
    """
    # Nested comments not handled.
    in_comment = False
    stripped_contents = ""
    for char in contents.lower():
        if char == "{":
            in_comment = True
        elif char == "}":
            in_comment = False
        elif not in_comment and not char.isspace():
            stripped_contents += char
    return stripped_contents.split(";")


def parse_chunk(value, elements, chunks):
    log.debug("Parse_chunk %s", value)
    chunk = []
    parts = utils.split_ignoring_parentheses(value, ",")
    log.debug(parts)
    for part in parts:
        if "symmetry" in part:
            continue
        if "inv" in part:
            # Reverse a sequence of elements. When doing this, swap
            # the entrance and exit angles for any dipoles.
            chunk_to_invert = re.match("inv\\((.*)\\)", part).groups()[0]
            inverted_chunk = []
            for el in reversed(chunks[chunk_to_invert]):
                if el.__class__ == Dipole:
                    inverted_dipole = el.copy()
                    inverted_dipole.EntranceAngle = el.ExitAngle
                    inverted_dipole.ExitAngle = el.EntranceAngle
                    inverted_chunk.append(inverted_dipole)
                else:
                    inverted_chunk.append(el)
            chunk.extend(inverted_chunk)
        elif "*" in part:
            num, contents = part.split("*")
            if contents.startswith("("):
                assert contents[-1] == ")"
                contents = contents[1:-1]
            chunk.extend(int(num) * parse_chunk(contents, elements, chunks))
        elif part in elements:
            chunk.append(elements[part])
        elif part in chunks:
            chunk.extend(chunks[part])
        else:
            raise Exception("part {} not understood".format(part))
    return chunk


def expand_tracy(contents, lattice_key, harmonic_number):
    lines = parse_lines(contents)
    assert lines[0] == "definelattice"
    assert lines[-2] == "end"
    assert lines[-1] == ""
    variables = {"harmonic_number": harmonic_number}
    elements = {}
    chunks = {}
    for line in lines[1:-2]:
        if ":" not in line:
            key, value = line.split("=")
            if key == "energy":
                value = parse_float(value, variables) * 1e9
            try:
                variables[key] = parse_float(value, variables)
            except ValueError:
                variables[key] = value
        else:
            key, value = line.split(":")
            if value.split(",")[0].strip() in ELEMENT_MAP:
                elements[key] = tracy_element_from_string(key, value, variables)
            else:
                chunk = parse_chunk(value, elements, chunks)
                chunks[key] = chunk

    return chunks[lattice_key], variables["energy"]


def parse_hom(hom_string, variables):
    """Parse 'hom' string from lattice file. Note that PolynomB
    is before PolynomA.

    hom(3, 1.2, 3.4) => [0, 0, 3.4], [0, 0, 1.2]
    """
    hom_parts = [part.strip() for part in hom_string.split(",")]
    assert len(hom_parts) % 3 == 0
    max_order = max(int(p) for i, p in enumerate(hom_parts) if i % 3 == 0)
    polynom_a = [0] * int(max_order)
    polynom_b = [0] * int(max_order)
    for i in range(len(hom_parts) // 3):
        order = int(hom_parts[i * 3])
        # PolynomB before PolynomA.
        b = parse_float(hom_parts[i * 3 + 1], variables)
        a = parse_float(hom_parts[i * 3 + 2], variables)
        polynom_a[order - 1] = a
        polynom_b[order - 1] = b

    return polynom_a, polynom_b


def tracy_element_from_string(name, element_string, variables):
    log.debug("Parsing tracy element {}".format(element_string))
    parts = utils.split_ignoring_parentheses(element_string, ",")
    params = {}
    element_type = parts[0]
    for part in parts[1:]:
        try:
            key, value = utils.split_ignoring_parentheses(part, "=")
        except ValueError:
            key, value = part, None
        if value in variables:
            value = variables[value]
        if key == "hom":
            assert value[0] == "("
            assert value[-1] == ")"
            polynom_a, polynom_b = parse_hom(value[1:-1], variables)
            params["PolynomA"] = polynom_a
            params["PolynomB"] = polynom_b
        else:
            params[key] = value

    return ELEMENT_MAP[element_type](name, params, variables)


def load_tracy(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Tracy file

    Parameters:
        filename:           Name of a Tracy file

    Keyword Args:
        use (str):          Name of the variable containing the desired lattice.
          Default: ``cell``
        name (str):         Name of the lattice. Default: taken from the file.
        energy (float):     Energy of the lattice [eV]
        periodicity(int):   Number of periods. Default: taken from the elements, or 1
        *:                  All other keywords will be set as Lattice attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    try:
        def elem_iterator(params, tracy_file):
            with open(params.setdefault("in_file", tracy_file)) as f:
                contents = f.read()
                element_lines, energy = expand_tracy(
                    contents, lattice_key, harmonic_number
                )
                params.setdefault("energy", energy)
                for line in element_lines:
                    yield line

        if "lattice_key" in kwargs:  # process the deprecated 'lattice_key' keyword
            kwargs.setdefault("use", kwargs.pop("lattice_key"))
        harmonic_number = kwargs.pop("harmonic_number")
        lattice_key = kwargs.pop("use", "cell")
        return Lattice(abspath(filename), iterator=elem_iterator, **kwargs)
    except Exception as e:
        raise ValueError("Failed to load tracy " "lattice {}: {}".format(filename, e))


register_format(".lat", load_tracy, descr="Tracy format. See :py:func:`.load_tracy`.")



"""
Functions relating to particle generation
"""
import numpy as np
from numpy.linalg import cholesky, LinAlgError
from warnings import warn
from ..lattice import AtError, AtWarning, Lattice, Orbit, random

__all__ = ['beam', 'sigma_matrix']


# noinspection PyPep8Naming
def sigma_matrix(ring: Lattice = None, **kwargs):
    r"""
    Calculate the correlation matrix :math:`\Sigma` to be used for particle
    generation

    Parameters:
        ring:           Lattice object or list of twiss parameters.

    Keyword Args:
        beam:               Particle coordinates
        twiss_in:           Data structure containing input twiss parameters.
        betax (float):      Input horizontal beta function [m]
        alphax (float):     Input horizontal alpha function [m]
        emitx (float):      Horizontal emittance [m.rad]
        betay (float):      Input vertical beta function [m]
        alphay (float):     Input vertical alpha function [m]
        emity (float):      Vertical emittance [m.rad]
        blength (float):    One sigma bunch length [m]
        espread (float):    One sigma energy spread [dp/p]

    Returns:
        sigma_matrix:       (6, 6) :math:`\Sigma`-matrix

    *beam* is a (6, nparticles) coordinates array of a particle population.
    If provided, all other parameters are ignored.

    If *twiss_in* is provided, its *R* or *alpha* and *beta* fields are used,
    *emitx* and *emity* must be provided.

    If *ring* is provided:

    * for a 6d lattice, :py:func:`.ohmi_envelope` is called to compute the
      equilibrium emittances. *emitx*, *emity* may overload these values.
    * for a 4d lattice, *emitx*, *emity* must be provided.

    Otherwise, *alphax*, *betax*, *emitx*, *alphay*, *betay* and *emity*
    must be provided. The result is an uncoupled :math:`\Sigma`-matrix

    If *espread* and *blength* are not specified, the results is a
    monochromatic beam.
    """
    def require(mess, *keys):
        miss = [key for key in keys if key not in kwargs]
        if len(miss) == 0:
            return [kwargs.pop(key) for key in keys]
        else:
            raise AtError('{0}: missing {1}'.format(mess, ', '.join(miss)))

    def ignore(mess):
        ok = kwargs.keys()
        if any(ok):
            warn(AtWarning('{0}: {1} ignored'.format(mess, ', '.join(ok))))

    def r_s():
        rs = np.zeros((2, 2))
        if any(('espread' in kwargs, 'blength' in kwargs)):
            blength, espread = require(
                'Both "blength" and "espread" are required',
                'blength', 'espread')
            ems = blength * espread
            if ems > 0.0:
                rs = np.array([[espread / blength, 0], [0, blength / espread]])
        else:
            ems = None
        return ems, rs

    def r_4d(ax, ay, bx, by, rs):
        r6 = np.zeros((3, 6, 6))
        r6[0, :2, :2] = np.array([[bx, -ax], [-ax, (1+ax**2)/bx]])
        r6[1, 2:4, 2:4] = np.array([[by, -ay], [-ay, (1+ay**2)/by]])
        r6[2, 4:, 4:] = rs
        return r6

    def r_expand(rmat, rs):
        if rmat.shape[0] < 3:
            r6 = np.zeros((3, 6, 6))
            r6[:2, :4, :4] = rmat
            r6[2, 4:, 4:] = rs
        else:
            r6 = rmat
        return r6

    def sigma_from_rmat(rmat, emit, embl):

        if emit is None:
            emx, emy = require('Emittances must be provided', 'emitx', 'emity')
            ems = embl
        else:
            emx = kwargs.pop('emitx', emit[0])
            emy = kwargs.pop('emity', emit[1])
            ems = emit[2] if embl is None else embl

        sigm = emx * rmat[0, :, :] + emy * rmat[1, :, :]
        if ems is None:
            warn(AtWarning('Monochromatic beam: no energy spread'))
        else:
            sigm += ems * rmat[2, :, :]
        return sigm

    if ring is not None:
        kwargs['ring'] = ring

    if 'beam' in kwargs:
        message = "'beam' is provided"
        beam = kwargs.pop('beam')
        orbit = kwargs.pop('orbit', None)
        if orbit is None:
            orbit = np.mean(beam, axis=1)
        beam -= orbit.reshape((6, 1))
        sigmat = (beam @ beam.T) / beam.shape[1]

    elif 'twiss_in' in kwargs:
        twin = kwargs.pop('twiss_in')
        message = "'twiss_in' is provided"
        em56, r56 = r_s()
        try:
            Rin = twin['R']
        except (ValueError, KeyError):  # record arrays throw ValueError !
            Rm = r_4d(twin['alpha'][0], twin['alpha'][1],
                      twin['beta'][0], twin['beta'][1], r56)
        else:
            Rm = r_expand(Rin, r56)
        sigmat = sigma_from_rmat(Rm, None, em56)

    elif 'ring' in kwargs:
        ring = kwargs.pop('ring')
        message = "'ring' is provided"
        em56, r56 = r_s()
        el0, _, _ = ring.linopt6(orbit=kwargs.pop('orbit', None))
        orbit = el0.closed_orbit
        Rm = r_expand(el0['R'], r56)
        if ring.is_6d:
            _, beam0, _ = ring.ohmi_envelope(orbit=orbit)
            sigmat = sigma_from_rmat(Rm, beam0.mode_emittances, em56)
        else:
            sigmat = sigma_from_rmat(Rm, None, em56)

    else:
        message = "Uncoupled beam"
        alphax, alphay, betax, betay = require(
            'Neither "twiss_in" nor "ring" is provided',
            'alphax', 'alphay', 'betax', 'betay')
        # orbit = kwargs.pop('orbit', np.zeros(6))
        em56, r56 = r_s()
        Rm = r_4d(alphax, alphay, betax, betay, r56)
        sigmat = sigma_from_rmat(Rm, None, em56)

    ignore(message)
    return sigmat


def beam(nparts: int, sigma, orbit: Orbit = None):
    r"""
    Generates an array of random particles according to the given
    :math:`\Sigma`-matrix

    Parameters:
        nparts:         Number of particles
        sigma:          :math:`\Sigma`-matrix as calculated by
          :py:func:`sigma_matrix`
        orbit:          An orbit can be provided to give a center of
          mass offset to the distribution

    Returns:
        particle_dist:  a (6, *nparts*) matrix of coordinates
    """
    def _get_single_plane(slc):
        try:
            return cholesky(sigma[slc, slc])
        except LinAlgError:
            return np.zeros((2, 2))

    v = random.thread.normal(size=(sigma.shape[1], nparts))

    try:
        # Try full 6x6 matrix
        lmat = cholesky(sigma)
        print("h, v, delta")
    except LinAlgError:
        lmat = np.zeros((6, 6))
        try:
            # Try x-y 4x4 matrix
            sel = range(4)
            idx = np.ix_(sel, sel)
            lmat[idx] = cholesky(sigma[idx])
            lmat[4:, 4:] = _get_single_plane(slice(4, 6))
            print("h, v")
        except LinAlgError:
            try:
                # Try x-delta 4x4 matrix
                sel = [0, 1, 4, 5]
                idx = np.ix_(sel, sel)
                lmat[idx] = cholesky(sigma[idx])
                lmat[2:4, 2:4] = _get_single_plane(slice(2, 4))
                print("h, delta")
            except LinAlgError:
                # Uncoupled
                lmat[:2, :2] = _get_single_plane(slice(2))
                lmat[2:4, 2:4] = _get_single_plane(slice(2, 4))
                lmat[4:, 4:] = _get_single_plane(slice(4, 6))
                print("uncoupled")

    particle_dist = np.asfortranarray(lmat @ v)

    if orbit is not None:
        particle_dist += orbit.reshape((6, 1))

    return particle_dist



"""
Functions relating to fast_ring
"""
from __future__ import annotations
from functools import reduce
import numpy
from typing import Union
from collections.abc import Sequence
from at.lattice import Lattice, Particle
from at.lattice import RFCavity, Element, Marker, get_cells, checkname
from at.lattice import get_elements, M66, SimpleQuantDiff, AtError, SimpleRadiation
from at.physics import gen_m66_elem, gen_detuning_elem, gen_quantdiff_elem
from at.constants import clight
import copy


__all__ = ['fast_ring', 'simple_ring']


def _rearrange(ring: Lattice, split_inds=[]):
    inds = numpy.append(split_inds, [0, len(ring)+1])
    inds = numpy.unique(inds)
    all_rings = [ring[int(b):int(e)] for b, e in zip(inds[:-1], inds[1:])]

    ringm = []
    for ring_slice in all_rings:
        ring_slice.insert(0, Marker('xbeg'))
        ring_slice.append(Marker('xend'))
        cavs = [e for e in ring_slice if isinstance(e, RFCavity)]
        newpass = ['IdentityPass' if c.Length == 0
                   else 'DriftPass' for c in cavs]
        for c, pm in zip(cavs, newpass):
            c.PassMethod = pm
        uni_freq = numpy.unique([e.Frequency for e in cavs])
        for fr in numpy.atleast_1d(uni_freq):
            cavf = [c for c in cavs if c.Frequency == fr]
            vol = reduce(lambda x, y: x+y, (c.Voltage for c in cavf))
            cavl = RFCavity('CAVL', 0, vol, fr,
                            cavf[0].HarmNumber, cavf[0].Energy)
            cavl.TimeLag = cavf[0].TimeLag
            ring_slice.append(cavl)
        ringm = ringm + ring_slice
    return all_rings, Lattice(ringm, energy=ring.energy)


def _fring(ring, split_inds=[], detuning_elem=None):
    all_rings, merged_ring = _rearrange(ring, split_inds=split_inds)
    ibegs = get_cells(merged_ring, checkname('xbeg'))
    iends = get_cells(merged_ring, checkname('xend'))
    _, orbit = merged_ring.find_orbit(refpts=ibegs | iends)
    if detuning_elem is None:
        detuning_elem = gen_detuning_elem(merged_ring, orbit[-1])
    else:
        detuning_elem.T1 = -orbit[-1]
        detuning_elem.T2 = orbit[-1]

    fastring = []
    for counter, r in enumerate(all_rings):
        cavs = [e for e in r if e.PassMethod.endswith('CavityPass')]
        [r.remove(c) for c in cavs]
        lin_elem = gen_m66_elem(r, orbit[2*counter],
                                orbit[2*counter+1])
        lin_elem.FamName = lin_elem.FamName + '_' + str(counter)
        [fastring.append(c) for c in cavs]
        fastring.append(lin_elem)
    fastring.append(detuning_elem)
    try:
        qd_elem = gen_quantdiff_elem(merged_ring)
        fastring.append(qd_elem)
    except ValueError:  # No synchrotron radiation => no diffusion element
        pass
    fastring = Lattice(fastring, **vars(ring))
    return fastring


def fast_ring(ring: Lattice, split_inds=[]) -> tuple[Lattice, Lattice]:
    """Generates a "fast ring"

    A fast ring consisting in:

    * a RF cavity per distinct frequency,
    * a 6x6 linear transfer map,
    * a detuning and chromaticity element,
    * a quantum diffusion element (for radiation ring).

    2 new lattices are returned, one with radiation and one without
    It is possible to split the original ring in multiple "fastrings"
    using the ``split_inds`` argument

    Parameters:
        ring:       Lattice description
        split_inds: List of indexes where to split the ring

    Returns:
        fring (Lattice):    Fast ring without radiation
        fringrad (Lattice): Fast ring with radiation
    """
    ringi = ring.deepcopy()
    fastringnorad = _fring(ringi.radiation_off(copy=True),
                           split_inds=split_inds)
    detuning_elem = copy.deepcopy(get_elements(fastringnorad,
                                               'NonLinear')[0])
    fastringrad = _fring(ringi.radiation_on(copy=True),
                         split_inds=split_inds,
                         detuning_elem=detuning_elem)
    return fastringnorad, fastringrad


def simple_ring(energy: float, circumference: float,
                harmonic_number: Union[float, Sequence[float]],
                Qx: float, Qy: float,
                Vrf: Union[float, Sequence[float]],
                alpha: float,
                betax: float = 1.0, betay: float = 1.0,
                alphax: float = 0.0, alphay: float = 0.0,
                dispx: float = 0.0, dispxp: float = 0.0,
                dispy: float = 0.0, dispyp: float = 0.0,
                Qpx: float = 0.0, Qpy: float = 0.0,
                A1: float = 0.0, A2: float = 0.0,
                A3: float = 0.0, emitx: float = 0.0,
                emity: float = 0.0, espread: float = 0.0,
                taux: float = 0.0, tauy: float = 0.0,
                tauz: float = 0.0, U0: float = 0.0,
                name: str = "",
                particle: Union[str, Particle] = 'relativistic',
                TimeLag: Union[float, Sequence[float]] = 0.0,
                ) -> Lattice:
    """Generates a "simple ring" based on a given dictionary
       of global parameters

    A simple ring consists of:

    * an RF cavity,
    * a 6x6 linear transfer map with no radiation damping,
    * a detuning and chromaticity element,
    * a simple radiation damping element
    * a simplified quantum diffusion element which contains equilibrium emittance

    Parameters:
        energy: [eV]
        circumference: [m]
        harmonic_number: can be scalar or sequence of scalars. The RF
          frequency is derived from this and the ring circumference
        Qx: horizontal tune
        Qy: vertical tune
        Vrf: RF Voltage set point [V] - can be scalar or sequence of scalars
        alpha: momentum compaction factor
        betax: horizontal beta function [m], Default=1
        betay: vertical beta function [m], Default=1
        alphax: horizontal alpha function, Default=0
        alphay: vertical alpha function, Default=0
        dispx: horizontal dispersion [m], Default=0
        dispxp: horizontal dispersion prime, Default=0
        dispy: vertical dispersion [m], Default=0
        dispyp: vertical dispersion prime, Default=0
        Qpx: horizontal linear chromaticity, Default=0
        Qpy: vertical linear chromaticity, Default=0
        A1: horizontal amplitude detuning coefficient, Default=0
        A2: cross term for amplitude detuning coefficient, Default=0
        A3: vertical amplitude detuning coefficient, Default=0
        emitx: horizontal equilibrium emittance [m.rad], Default=0
          ignored if emitx=0
        emity: vertical equilibrium emittance [m.rad], Default=0
          ignored if emity=0
        espread: equilibrium momentum spread, Default=0
          ignored if espread=0
        taux: horizontal radiation damping time [turns], Default=0
          ignored if taux=0
        tauy: vertical radiation damping time [turns], Default=0
          ignored if tauy=0
        tauz: longitudinal radiation damping time [turns], Default=0
          ignored if tauz=0
        U0: energy loss [eV] (positive number), Default=0
        name: Name of the lattice
        particle: circulating particle. May be
          'relativistic', 'electron', 'positron', 'proton'
          or a Particle object
        TimeLag: Set the timelag of the cavities, Default=0. Can be scalar
          or sequence of scalars (as with harmonic_number and Vrf).

    If the given emitx, emity or espread is 0, then no equlibrium emittance
    is applied in this plane.
    If the given tau is 0, then no radiation damping is applied for this plane.

    Returns:
        ring:    Simple ring
    """
    try:
        rfp = numpy.broadcast(Vrf, harmonic_number, TimeLag)
    except ValueError:
        raise AtError('Vrf, harmonic_number and TimeLag must be broadcastable')

    # revolution frequency
    f0 = clight / circumference

    all_cavities = [RFCavity(f"RFC{i+1}", 0.0, v, h*f0, h, energy, TimeLag=t)
                    for i, (v, h, t) in enumerate(rfp)]

    # Now we will use the optics parameters to compute the uncoupled M66 matrix

    s_dphi_x = numpy.sin(2*numpy.pi*Qx)
    c_dphi_x = numpy.cos(2*numpy.pi*Qx)
    s_dphi_y = numpy.sin(2*numpy.pi*Qy)
    c_dphi_y = numpy.cos(2*numpy.pi*Qy)

    M00 = c_dphi_x + alphax * s_dphi_x
    M01 = betax * s_dphi_x
    M10 = -(1. + alphax**2) / betax * s_dphi_x
    M11 = c_dphi_x - alphax * s_dphi_x

    M04 = (1 - M00) * dispx - M01 * dispxp
    M14 = -M10 * dispx + (1 - M11) * dispxp
    
    M22 = c_dphi_y + alphay * s_dphi_y
    M23 = betay * s_dphi_y
    M32 = -(1. + alphay**2) / betay * s_dphi_y
    M33 = c_dphi_y - alphay * s_dphi_y

    M24 = (1 - M22) * dispy - M23 * dispyp
    M34 = -M32 * dispy + (1 - M33) * dispyp
    
    
    M44 = 1.
    M45 = 0.
    M54 = alpha*circumference
    M55 = 1

    Mat66 = numpy.array([[M00, M01, 0., 0., M04, 0.],
                         [M10, M11, 0., 0., M14, 0.],
                         [0., 0., M22, M23, M24, 0.],
                         [0., 0., M32, M33, M34, 0.],
                         [0., 0., 0., 0., M44, M45],
                         [0., 0., 0., 0., M54, M55]], order='F')

    # generate the linear tracking element, we set a length
    # which is needed to give the lattice object the correct length
    # (although it is not used for anything else)
    lin_elem = M66('Linear', m66=Mat66, Length=circumference)

    # Generate the simple radiation element
    simplerad = SimpleRadiation('SR', taux=taux, tauy=tauy, 
                                tauz=tauz, U0=U0, dispx=dispx,
                                dispy=dispy, dispxp=dispxp, 
                                dispyp=dispyp)
                                
    # Generate the simple quantum diffusion element
    quantdiff = SimpleQuantDiff('SQD', betax=betax, betay=betay,
                                emitx=emitx, emity=emity,
                                espread=espread, taux=taux,
                                tauy=tauy, tauz=tauz)

    # Generate the detuning element
    nonlin_elem = Element('NonLinear', PassMethod='DeltaQPass',
                          Betax=betax, Betay=betay,
                          Alphax=alphax, Alphay=alphay,
                          Qpx=Qpx, Qpy=Qpy,
                          A1=A1, A2=A2, A3=A3)

    # Assemble all elements into the lattice object
    ring = Lattice(all_cavities + [lin_elem, nonlin_elem, simplerad, quantdiff],
                   name=name, energy=energy, particle=particle, periodicity=1)

    return ring



import at
from at.lattice.elements import Corrector, Dipole, Quadrupole, ThinMultipole
from math import pi
from numpy.testing import assert_allclose as assert_close
import pytest


def simple_fodo(ncells: int = 64) -> at.Lattice:
    """Return a simple F0D0 lattice"""
    # dipole
    dipole_length = 1.0
    bending_angle = pi / ncells
    # Focusing quad
    kf = 0.5
    lf = 1.0
    # Defocusing quad
    kd = -0.5
    ld = 1.0
    # Drifts
    drift_f = 0.5
    drift_d = 0.5 + 0.5 * ld

    drf = at.Drift('drf', drift_f)
    drd = at.Drift('drd', drift_d)
    qf0 = Quadrupole('qf0', lf, kf)
    qd0 = ThinMultipole('qd0', [0, 0], [0, kd])
    dip0 = Dipole('dip0', dipole_length, bending_angle, 0)
    kick0 = Corrector('kick0', 0.0, [0.001, -0.001])
    # build the nominal lattice
    fodo0 = at.Lattice([drd, kick0, dip0, drf, qf0, drf, dip0, drd, qd0],
                       energy=6.0E9, name='fodo0')
    return fodo0


@pytest.mark.parametrize('dp', [0.01, -0.01])
@pytest.mark.parametrize('ncells', [64, 6])
def test_scaling(dp, ncells):
    """Check the FieldScaling attribute"""
    scaling = 1.0 + dp
    fodo0 = simple_fodo(ncells)
    magnets = fodo0.get_bool_index(
        at.checktype((Corrector, Dipole, Quadrupole, ThinMultipole)))

    # Nominal optics:
    _, rd0, el0 = at.get_optics(fodo0, refpts=at.All)

    # Off-momentum optics with scaled magnets
    fodoa = fodo0.replace(magnets)
    for elem in fodoa[magnets]:
        elem.FieldScaling = scaling
    _, rda, ela = at.get_optics(fodoa, dp=dp, refpts=at.All)
    assert_close(rd0.tune, rda.tune, atol=1.e-10, rtol=0)
    assert_close(el0.closed_orbit[:, [0, 2]],
                 ela.closed_orbit[:, [0, 2]], atol=1.e-10, rtol=0)

    # Off-momentum optics with scaling the whole lattice
    scin = at.Element('scin', PassMethod='ChangePRefPass',
                      FieldScaling=scaling)
    scout = at.Element('scout', PassMethod='ChangePRefPass',
                       FieldScaling=1.0 / scaling)
    fodob = fodo0.copy()
    fodob.insert(0, scin)
    fodob.append(scout)
    _, rdb, elb = at.get_optics(fodob, dp=dp, refpts=at.All)
    assert_close(rd0.tune, rdb.tune, atol=1.e-10, rtol=0)
    assert_close(el0.closed_orbit[:,    [0, 2]],
                 elb.closed_orbit[1:-1, [0, 2]], atol=1.e-10, rtol=0)



import numpy
from .elements import Element, _array
from .utils import AtError
from typing import Optional, Union
from enum import IntEnum
from datetime import datetime


class ACMode(IntEnum):
    """Class to define the excitation types"""
    SINE = 0
    WHITENOISE = 1
    ARBITRARY = 2


class VariableMultipole(Element):
    """Class to generate an AT variable thin multipole element
    """
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES
    _conversions = dict(Element._conversions,
                        Mode=int,
                        AmplitudeA=_array, AmplitudeB=_array,
                        FrequencyA=float, FrequencyB=float,
                        PhaseA=float, PhaseB=float,
                        Seed=int, NSamplesA=int, NSamplesB=int,
                        FuncA=_array, FuncB=_array, Ramps=_array,
                        Periodic=bool)

    def __init__(self, family_name, AmplitudeA=None, AmplitudeB=None,
                 mode=ACMode.SINE, **kwargs):
        # noinspection PyUnresolvedReferences
        r"""
        Parameters:
            family_name(str):    Element name

        Keyword Arguments:
            AmplitudeA(list,float): Amplitude of the excitation for PolynomA. Default None
            AmplitudeB(list,float): Amplitude of the excitation for PolynomB. Default None
            mode(ACMode): defines the evaluation grid. Default ACMode.SINE

              * :py:attr:`.ACMode.SINE`: sine function
              * :py:attr:`.ACMode.WHITENOISE`: gaussian white noise
              * :py:attr:`.GridMode.ARBITRARY`: user defined turn-by-turn kick list
            FrequencyA(float): Frequency of the sine excitation for PolynomA
            FrequencyB(float): Frequency of the sine excitation for PolynomB
            PhaseA(float): Phase of the sine excitation for PolynomA. Default 0
            PhaseB(float): Phase of the sine excitation for PolynomB. Default 0
            MaxOrder(int): Order of the multipole for scalar amplitude. Default 0
            Seed(int): Seed of the random number generator for white
                       noise excitation. Default datetime.now()
            FuncA(list): User defined tbt kick list for PolynomA
            FuncB(list): User defined tbt kick list for PolynomB
            Periodic(bool): If True (default) the user defined kick is repeated
            Ramps(list): Vector (t0, t1, t2, t3) in turn number to define the ramping 
                         of the excitation

              * ``t<t0``: excitation amplitude is zero
              * ``t0<t<t1``: exciation amplitude is linearly ramped up
              * ``t1<t<t2``: exciation amplitude is constant             
              * ``t2<t<t3``: exciation amplitude is linearly ramped down
              * ``t3<t``: exciation amplitude is zero           

        Examples:

            >>> acmpole = at.VariableMultipole('ACMPOLE', AmplitudeB=amp, FrequencyB=frequency)
            >>> acmpole = at.VariableMultipole('ACMPOLE', AmplitudeB=amp, mode=at.ACMode.WHITENOISE)
            >>> acmpole = at.VariableMultipole('ACMPOLE', AmplitudeB=amp, FuncB=fun, mode=at.ACMode.ARBITRARY)

        .. note::

            * For all excitation modes all least one amplitude has to be provided.
              The default excitation is ``ACMode.SINE``
            * For ``mode=ACMode.SINE`` the ``Frequency(A,B)`` corresponding to the
              ``Amplitude(A,B)`` has to be provided
            * For ``mode=ACMode.ARBITRARY`` the ``Func(A,B)`` corresponding to the
              ``Amplitude(A,B)`` has to be provided
        """
        kwargs.setdefault('PassMethod', 'VariableThinMPolePass')
        self.MaxOrder = kwargs.pop('MaxOrder', 0)
        self.Periodic = kwargs.pop('Periodic', True)
        self.Mode = int(mode)
        if AmplitudeA is None and AmplitudeB is None:
            raise AtError('Please provide at least one amplitude for A or B')
        AmplitudeB = self._set_params(AmplitudeB, mode, 'B', **kwargs)
        AmplitudeA = self._set_params(AmplitudeA, mode, 'A', **kwargs)
        self._setmaxorder(AmplitudeA, AmplitudeB)
        if mode == ACMode.WHITENOISE:
            self.Seed = kwargs.pop('Seed', datetime.now().timestamp())
        self.PolynomA = numpy.zeros(self.MaxOrder+1)
        self.PolynomB = numpy.zeros(self.MaxOrder+1)
        ramps = kwargs.pop('Ramps', None)
        if ramps is not None:
            assert len(ramps)==4, \
                'Ramps has to be a vector with 4 elements'
            self.Ramps = ramps
        super(VariableMultipole, self).__init__(family_name, **kwargs)

    def _setmaxorder(self, ampa, ampb):
        mxa, mxb = 0, 0
        if ampa is not None:
            mxa = numpy.max(numpy.nonzero(ampa))
        if ampb is not None:
            mxb = numpy.max(numpy.nonzero(ampb))
        self.MaxOrder = max(mxa, mxb)
        if ampa is not None:
            delta = self.MaxOrder - len(ampa)
            if delta > 0:
                ampa = numpy.pad(ampa, (0, delta))
            setattr(self, 'AmplitudeA', ampa)
        if ampb is not None:
            delta = self.MaxOrder + 1 - len(ampb)
            if delta > 0:
                ampb = numpy.pad(ampb, (0, delta))
            setattr(self, 'AmplitudeB', ampb)

    def _set_params(self, amplitude, mode, ab, **kwargs):
        if amplitude is not None:
            if numpy.isscalar(amplitude):
                amp = numpy.zeros(self.MaxOrder)
                amplitude = numpy.append(amp, amplitude)
            if mode == ACMode.SINE:
                self._set_sine(ab, **kwargs)
            if mode == ACMode.ARBITRARY:
                self._set_arb(ab, **kwargs)
        return amplitude

    def _set_sine(self, ab, **kwargs):
        frequency = kwargs.pop('Frequency' + ab, None)
        phase = kwargs.pop('Phase' + ab, 0)
        assert frequency is not None, \
            'Please provide a value for Frequency' + ab
        setattr(self, 'Frequency' + ab, frequency)
        setattr(self, 'Phase' + ab, phase)

    def _set_arb(self, ab, **kwargs):
        func = kwargs.pop('Func' + ab, None)
        nsamp = len(func)
        assert func is not None, \
            'Please provide a value for Func' + ab
        setattr(self, 'Func' + ab, func)
        setattr(self, 'NSamples' + ab, nsamp)



"""
Radiation and equilibrium emittances
"""
from __future__ import annotations
from math import sin, cos, tan, sqrt, sinh, cosh, pi
import numpy
from typing import Union
from scipy.linalg import inv, det, solve_sylvester
from at.lattice import Lattice, check_radiation, Refpts, All
from at.lattice import Dipole, Wiggler, DConstant
from at.lattice import Quadrupole, Multipole, QuantumDiffusion
from at.lattice import frequency_control, set_value_refpts
from at.tracking import internal_lpass
from at.physics import find_orbit6, find_m66, find_elem_m66, Orbit
from at.physics import find_mpole_raddiff_matrix, get_tunes_damp
from at.physics import ELossMethod

__all__ = ['ohmi_envelope', 'get_radiation_integrals', 'quantdiffmat',
           'gen_quantdiff_elem', 'tapering']

_NSTEP = 60  # nb slices in a wiggler period

_submat = [slice(0, 2), slice(2, 4), slice(6, 3, -1)]

# dtype for structured array containing optical parameters
ENVELOPE_DTYPE = [('r66', numpy.float64, (6, 6)),
                  ('r44', numpy.float64, (4, 4)),
                  ('m66', numpy.float64, (6, 6)),
                  ('orbit6', numpy.float64, (6,)),
                  ('emitXY', numpy.float64, (2,)),
                  ('emitXYZ', numpy.float64, (3,))]


def _dmatr(ring: Lattice, orbit: Orbit = None, keep_lattice: bool = False):
    """
    compute the cumulative diffusion and orbit
    matrices over the ring
    """

    def _cumulb(it):
        """accumulate diffusion matrices"""
        cumul = numpy.zeros((6, 6))
        yield cumul
        for el, orbin, b in it:
            m = find_elem_m66(el, orbin, energy=energy, particle=ring.particle)
            cumul = m.dot(cumul).dot(m.T) + b
            yield cumul

    energy = ring.energy

    if orbit is None:
        orbit, _ = find_orbit6(ring, keep_lattice=keep_lattice)
        keep_lattice = True

    orbs = numpy.squeeze(
        internal_lpass(ring, orbit.copy(order='K'), refpts=All,
                       keep_lattice=keep_lattice), axis=(1, 3)).T
    b0 = numpy.zeros((6, 6))
    bb = [find_mpole_raddiff_matrix(elem, elemorb, energy)
          if elem.PassMethod.endswith('RadPass') else b0
          for elem, elemorb in zip(ring, orbs)]
    bbcum = numpy.stack(list(_cumulb(zip(ring, orbs, bb))), axis=0)
    return bbcum, orbs


def _lmat(dmat):
    """
    lmat is Cholesky decomp of dmat unless diffusion is 0 in
    vertical.  Then do chol on 4x4 hor-long matrix and put 0's
    in vertical
    """
    lmat = numpy.zeros((6, 6))
    try:
        lmat = numpy.linalg.cholesky(dmat)
    except numpy.linalg.LinAlgError:
        nz = numpy.where(dmat != 0)
        cmat = numpy.reshape(dmat[nz], (4, 4))
        cmat = numpy.linalg.cholesky(cmat)
        lmat[nz] = numpy.reshape(cmat, (16,))
    return lmat


@check_radiation(True)
def ohmi_envelope(ring: Lattice, refpts: Refpts = None, orbit: Orbit = None,
                  keep_lattice: bool = False):
    """Calculates the equilibrium beam envelope

    Computation based on Ohmi's beam envelope formalism [1]_

    Parameters:
        ring:           Lattice description. Radiation must be ON
        refpts:         Observation points
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: False

    Returns:
        emit0 (numpy.recarray):     Emittance data at the start/end of the ring
        beamdata (numpy.recarray):  Beam parameters at the start of the ring
        emit (numpy.recarray):      Emittance data at the points selected to by
          ``refpts``

    **emit** is a :py:obj:`record array <numpy.recarray>` with the following
    fields:

    ================    ===================================================
    **r66**             (6, 6) equilibrium envelope matrix R
    **r44**             (4, 4) betatron emittance matrix (dpp = 0)
    **m66**             (6, 6) transfer matrix from the start of the ring
    **orbit6**          (6,) closed orbit
    **emitXY**          (2,) betatron emittance projected on xxp and yyp
    **emitXYZ**         (3,) 6x6 emittance projected on xxp, yyp, ldp
    ================    ===================================================

    Values given at the entrance of each element specified in ``refpts``.

    Field values can be obtained with either
    ``emit['r66']`` or ``emit.r66``

    **beamdata** is a :py:obj:`record array <numpy.recarray>` with the
    following fields:

    ====================  ===================================================
    **tunes**             tunes of the 3 normal modes
    **damping_rates**     damping rates of the 3 normal modes
    **mode_matrices**     R-matrices of the 3 normal modes
    **mode_emittances**   equilibrium emittances of the 3 normal modes
    ====================  ===================================================

    References:
        .. [1] K.Ohmi et al. Phys.Rev.E. Vol.49. (1994)
    """

    def process(r66):
        # projections on xx', zz', ldp
        emit3sq = numpy.array([det(r66[s, s]) for s in _submat])
        # Prevent from unrealistic negative values of the determinant
        emit3 = numpy.sqrt(numpy.maximum(emit3sq, 0.0))
        # Emittance cut for dpp=0
        if emit3[0] < 1.E-13:  # No equilibrium emittance
            r44 = numpy.nan * numpy.ones((4, 4))
        elif emit3[1] < 1.E-13:  # Uncoupled machine
            minv = inv(r66[[0, 1, 4, 5], :][:, [0, 1, 4, 5]])
            r44 = numpy.zeros((4, 4))
            r44[:2, :2] = inv(minv[:2, :2])
        else:  # Coupled machine
            minv = inv(r66)
            r44 = inv(minv[:4, :4])
        # betatron emittances (dpp=0)
        emit2sq = numpy.array(
            [det(r44[s, s], check_finite=False) for s in _submat[:2]])
        # Prevent from unrealistic negative values of the determinant
        emit2 = numpy.sqrt(numpy.maximum(emit2sq, 0.0))
        return r44, emit2, emit3

    def propag(m, cumb, orbit6):
        """Propagate the beam matrix to refpts"""
        sigmatrix = m.dot(rr).dot(m.T) + cumb
        m44, emit2, emit3 = process(sigmatrix)
        return sigmatrix, m44, m, orbit6, emit2, emit3

    uint32refs = ring.get_uint32_index(refpts)
    bbcum, orbs = _dmatr(ring, orbit=orbit, keep_lattice=keep_lattice)
    mring, ms = find_m66(ring, uint32refs, orbit=orbs[0], keep_lattice=True)
    # ------------------------------------------------------------------------
    # Equation for the moment matrix R is
    #         R = MRING*R*MRING' + BCUM;
    # We rewrite it in the form of Lyapunov-Sylvester equation to use scipy's
    # solve_sylvester function
    #            A*R + R*B = Q
    # where
    #               A =  inv(MRING)
    #               B = -MRING'
    #               Q = inv(MRING)*BCUM
    # ------------------------------------------------------------------------
    aa = inv(mring)
    bb = -mring.T
    qq = numpy.dot(aa, bbcum[-1])
    rr = solve_sylvester(aa, bb, qq)
    rr = 0.5 * (rr + rr.T)
    rr4, emitxy, emitxyz = process(rr)
    r66data = get_tunes_damp(mring, rr)

    data0 = numpy.rec.fromarrays(
        (rr, rr4, mring, orbs[0], emitxy, emitxyz),
        dtype=ENVELOPE_DTYPE)
    if uint32refs.shape == (0,):
        data = numpy.recarray((0,), dtype=ENVELOPE_DTYPE)
    else:
        data = numpy.rec.fromrecords(
            list(map(propag, ms, bbcum[uint32refs], orbs[uint32refs, :])),
            dtype=ENVELOPE_DTYPE)

    return data0, r66data, data


@frequency_control
def get_radiation_integrals(ring, dp: float = None, twiss=None, **kwargs)\
        -> tuple[float, float, float, float, float]:
    r"""Computes the 5 radiation integrals for uncoupled lattices.

    Parameters:
        ring:   Lattice description
        twiss:  Linear optics at all points (from :py:func:`.linopt6`).
          If ``None``, it will be computed.

    Keyword Args:
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to       
        method (Callable):  Method for linear optics:

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis

    Returns:
        i1 (float): Radiation integrals - :math:`I_1 \quad [m]`
        i2 (float): :math:`I_2 \quad [m^{-1}]`
        i3 (float): :math:`I_3 \quad [m^{-2}]`
        i4 (float): :math:`I_4 \quad [m^{-1}]`
        i5 (float): :math:`I_5 \quad [m^{-1}]`
    """

    def element_radiation(elem: Union[Dipole, Quadrupole], vini, vend):
        """Analytically compute the radiation integrals in dipoles"""
        beta0 = vini.beta[0]
        alpha0 = vini.alpha[0]
        eta0 = vini.dispersion[0]
        etap0 = vini.dispersion[1]
        theta = getattr(elem, 'BendingAngle', None)
        if theta is None:
            xpi = vini.closed_orbit[1]/(1+vini.closed_orbit[4])
            xpo = vend.closed_orbit[1]/(1+vend.closed_orbit[4])
            theta = xpi-xpo
        if abs(theta) < 1.0e-7:
            return numpy.zeros(5)
        angin = getattr(elem, 'EntranceAngle', 0.0)
        angout = getattr(elem, 'ExitAngle', 0.0)

        ll = elem.Length
        rho = ll / theta
        rho2 = rho * rho
        k2 = elem.K + 1.0 / rho2
        eps1 = tan(angin) / rho
        eps2 = tan(angout) / rho
        eta3 = vend.dispersion[0]
        alpha1 = alpha0 - beta0 * eps1
        gamma1 = (1.0 + alpha1 * alpha1) / beta0
        etap1 = etap0 + eta0 * eps1
        etap2 = vend.dispersion[1] - eta3 * eps2

        h0 = gamma1*eta0*eta0 + 2.0*alpha1*eta0*etap1 + beta0*etap1*etap1

        if k2 != 0.0:
            if k2 > 0.0:  # Focusing
                kl = ll * sqrt(k2)
                ss = sin(kl) / kl
                cc = cos(kl)
            else:  # Defocusing
                kl = ll * sqrt(-k2)
                ss = sinh(kl) / kl
                cc = cosh(kl)
            eta_ave = (theta - (etap2 - etap1)) / k2 / ll
            bb = 2.0 * (alpha1 * eta0 + beta0 * etap1) * rho
            aa = -2.0 * (alpha1 * etap1 + gamma1 * eta0) * rho
            h_ave = h0 + (aa * (1.0 - ss) + bb * (1.0 - cc) / ll
                          + gamma1 * (3.0 - 4.0 * ss + ss * cc) / 2.0 / k2
                          - alpha1 * (1.0 - cc) ** 2 / k2 / ll
                          + beta0 * (1.0 - ss * cc) / 2.0
                          ) / k2 / rho2
        else:
            eta_ave = 0.5 * (eta0 + eta3) - ll * ll / 12.0 / rho
            hp0 = 2.0 * (alpha1 * eta0 + beta0 * etap1) / rho
            h2p0 = 2.0 * (-alpha1 * etap1 + beta0 / rho - gamma1 * eta0) / rho
            h_ave = h0 + hp0 * ll / 2.0 + h2p0 * ll * ll / 6.0 \
                - alpha1 * ll ** 3 / 4.0 / rho2 \
                + gamma1 * ll ** 4 / 20.0 / rho2

        di1 = eta_ave * ll / rho
        di2 = ll / rho2
        di3 = ll / abs(rho) / rho2
        di4 = eta_ave * ll * (2.0 * elem.K + 1.0 / rho2) / rho \
            - (eta0 * eps1 + eta3 * eps2) / rho
        di5 = h_ave * ll / abs(rho) / rho2
        return numpy.array([di1, di2, di3, di4, di5])

    def wiggler_radiation(elem: Wiggler, dini):
        """Compute the radiation integrals in wigglers with the following
        approximations:

        - The wiggler is aligned with the closed orbit
        - The self-induced dispersion is neglected in I4 and I5, but is is used
          as a lower limit for the I5 contribution
        - I1, I2 are integrated analytically
        - I3 is integrated analytically for a single harmonic, numerically
          otherwise
        """

        def b_on_axis(wiggler: Wiggler, s):
            """On-axis wiggler field"""

            def harm(coef, h, phi):
                return -Bmax * coef * numpy.cos(h*kws + phi)

            kw = 2 * pi / wiggler.Lw
            Bmax = wiggler.Bmax
            kws = kw * s
            zz = [numpy.zeros(kws.shape)]
            vh = zz + [harm(pb[1], pb[4], pb[5]) for pb in wiggler.By.T]
            vv = zz + [-harm(pb[1], pb[4], pb[5]) for pb in wiggler.Bx.T]
            bys = numpy.sum(numpy.stack(vh), axis=0)
            bxs = numpy.sum(numpy.stack(vv), axis=0)
            return bxs, bys

        le = elem.Length
        alphax0 = dini.alpha[0]
        betax0 = dini.beta[0]
        gammax0 = (alphax0 * alphax0 + 1) / betax0
        eta0 = dini.dispersion[0]
        etap0 = dini.dispersion[1]
        H0 = gammax0*eta0*eta0 + 2*alphax0*eta0*etap0 + betax0*etap0*etap0
        avebetax = betax0 + alphax0*le + gammax0*le*le/3

        kw = 2 * pi / elem.Lw
        rhoinv = elem.Bmax / Brho
        coefh = elem.By[1, :] * rhoinv
        coefv = elem.Bx[1, :] * rhoinv
        coef2 = numpy.concatenate((coefh, coefv))
        if len(coef2) == 1:
            di3 = le * coef2[0] ** 3 * 4 / 3 / pi
        else:
            bx, bz = b_on_axis(elem, numpy.linspace(0, elem.Lw, _NSTEP + 1))
            rinv = numpy.sqrt(bx*bx + bz*bz) / Brho
            di3 = numpy.trapz(rinv ** 3) * le / _NSTEP
        di2 = le * (numpy.sum(coefh * coefh) + numpy.sum(coefv * coefv)) / 2
        di1 = -di2 / kw / kw
        di4 = 0
        if len(coefh) > 0:
            d5lim = 4 * avebetax * le * coefh[0] ** 5 / 15 / pi / kw / kw
        else:
            d5lim = 0
        di5 = max(H0 * di3, d5lim)
        return numpy.array([di1, di2, di3, di4, di5])

    Brho = ring.BRho
    integrals = numpy.zeros((5,))

    if twiss is None:
        _, _, twiss = ring.get_optics(refpts=range(len(ring) + 1), dp=dp,
                                      get_chrom=True, **kwargs)
    elif len(twiss) != len(ring) + 1:
        raise ValueError('length of Twiss data should be {0}'
                         .format(len(ring) + 1))
    for (el, vini, vend) in zip(ring, twiss[:-1], twiss[1:]):
        if isinstance(el, (Dipole, Quadrupole)):
            integrals += element_radiation(el, vini, vend)
        elif isinstance(el, Wiggler) and el.PassMethod != 'DriftPass':
            integrals += wiggler_radiation(el, vini)
    return tuple(integrals)


@check_radiation(True)
def quantdiffmat(ring: Lattice, orbit: Orbit = None) -> numpy.ndarray:
    """Computes the diffusion matrix of the whole ring

    Parameters:
        ring:           Lattice description. Radiation must be ON
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).

    Returns:
        diffmat (ndarray):  Diffusion matrix (6,6)
    """
    bbcum, _ = _dmatr(ring, orbit=orbit)
    diffmat = [(bbc + bbc.T) / 2 for bbc in bbcum]
    return numpy.round(diffmat[-1], 24)


@check_radiation(True)
def gen_quantdiff_elem(ring: Lattice, orbit: Orbit = None) -> QuantumDiffusion:
    """Generates a quantum diffusion element

    Parameters:
        ring:           Lattice description. Radiation must be ON
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).

    Returns:
        diffElem (QuantumDiffusion): Quantum diffusion element
    """
    dmat = quantdiffmat(ring, orbit=orbit)
    lmat = numpy.asfortranarray(_lmat(dmat))
    diff_elem = QuantumDiffusion('Diffusion', lmat)
    return diff_elem


@check_radiation(True)
def tapering(ring: Lattice, multipoles: bool = True,
             niter: int = 1, **kwargs) -> None:
    """Scales magnet strengths

    Scales magnet strengths with local energy to cancel the closed orbit
    and optics errors due to synchrotron radiations. The dipole bending angle
    is changed by changing the reference momentum. This is the ideal
    tapering scheme where magnets and multipoles components (PolynomB and
    PolynomA) are scaled individually.

    Warning:
        This method works only for lattices without errors and
        corrections: if not all corrections and field errors will also be
        scaled !!!

    Parameters:
        ring:           Lattice description
        multipoles:     Scales all multipoles
        niter:          Number of iteration

    Keyword Args:
        method (ELossMethod):   Method for energy loss computation.
          See :py:class:`.ELossMethod`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
    """

    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    method = kwargs.pop('method', ELossMethod.TRACKING)
    dipin = ring.get_bool_index(Dipole)
    dipout = numpy.roll(dipin, 1)
    multin = ring.get_bool_index(Multipole) & ~dipin
    multout = numpy.roll(multin, 1)

    for i in range(niter):
        _, o6 = find_orbit6(ring, refpts=range(len(ring)+1),
                            XYStep=xy_step, DPStep=dp_step, method=method)
        dpps = (o6[dipin, 4] + o6[dipout, 4]) / 2.0
        set_value_refpts(ring, dipin, 'FieldScaling', 1+dpps)
        if multipoles:
            _, o6 = find_orbit6(ring, refpts=range(len(ring)+1),
                                XYStep=xy_step, DPStep=dp_step, method=method)
            dppm = (o6[multin, 4] + o6[multout, 4]) / 2
            set_value_refpts(ring, multin, 'FieldScaling', 1+dppm)


Lattice.ohmi_envelope = ohmi_envelope
Lattice.get_radiation_integrals = get_radiation_integrals
Lattice.tapering = tapering



import at
from numpy.testing import assert_allclose as assert_close
import pytest
import numpy as np


@pytest.fixture(scope='session')
def test_ring(hmba_lattice):
    ring = hmba_lattice.deepcopy()
    sf = at.get_refpts(ring, 'SF*')
    sf1 = ring[sf[0]].divide([0.5, 0.5])
    sf2 = ring[sf[1]].divide([0.5, 0.5])
    ring.pop(sf[1])
    ring.insert(sf[1], sf2[1])
    ring.insert(sf[1], sf2[0])
    ring.pop(sf[0])
    ring.insert(sf[0], sf1[1])
    ring.insert(sf[0], sf1[0])
    return ring


def test_linopt_matching(test_ring):
    # Define the location of constraint
    sf = at.get_refpts(test_ring, 'SF*')[1::2]
    center = at.get_refpts(test_ring, 'CellCenter')

    # Define the variables
    names = ['QF1*', 'QD2*', 'QD3*', 'QF4*', 'QD5*', 'QF6*', 'QF8*']
    bounds = [[0, 5], [-5, 0], [-5, 0], [0, 5], [-5, 0], [0, 5], [0, 5]]
    variables = [at.ElementVariable(at.get_refpts(test_ring, nm), 'PolynomB',
                                    index=1, name=nm)
                 for nm, bnd in zip(names, bounds)]

    # Define an evaluation function for the phase advance (both planes)
    # noinspection PyUnusedLocal
    def mu_diff(lindata, tune, chrom):
        delta_mu = (lindata[1].mu - lindata[0].mu) / (2 * np.pi)
        return delta_mu % 1.0

    # Define the constraints
    cst1 = at.LinoptConstraints(test_ring, method=at.linopt2)
    cst1.add('tunes', [0.38, 0.85], name='tunes')
    # Start of the ring
    cst1.add('alpha', [0.0, 0.0], refpts=0, name='alpha_0')
    cst1.add('dispersion', 0.0, refpts=0, index=1, name="eta'_0")
    # Center of the cell
    cst1.add('beta', 4.69999257, refpts=center, index=1, name='betay_c')
    cst1.add('alpha', [0.0, 0.0], refpts=center, name='alpha_c')
    cst1.add('dispersion', 0.0, refpts=center, index=1, name="eta'_c")
    # Focusing sextupoles
    cst1.add('beta', [5.40000677, 5.39998124], refpts=sf, index=1,
             name='betay_sf')
    cst1.add('alpha', [0.68000392, -0.67999686], refpts=sf, index=1,
             name='alphay_sf')
    cst1.add('dispersion', [0.08820002, 0.08819999], refpts=sf, index=0,
             name="eta_sf")
    # Phase advance
    cst1.add(mu_diff, [0.49721338, 0.48228011], refpts=sf, name='delta phi')

    # Perform the fit
    newring = at.match(test_ring, variables, (cst1,), method='lm', verbose=1)

    # check the residuals
    assert_close(cst1.evaluate(newring), 0, rtol=0.0, atol=1e-4)


def test_envelope_matching(test_ring):
    # Define the variables
    test_ring = test_ring.radiation_on(copy=True)
    names = ['QF1*', 'QD2*']
    variables = [at.ElementVariable(at.get_refpts(test_ring, nm), 'PolynomB',
                                    index=1, name=nm) for nm in names]
    variables += [at.ElementVariable([0], 'Voltage', name='voltage')]
    # [0] instead of 0 because of bug in utils.py

    # Define the constraints
    lopcst = at.EnvelopeConstraints(test_ring)
    lopcst.add('tunes', [0.38, 0.85, 1.1e-4])

    # Perform the fit
    newring = at.match(test_ring, variables, (lopcst,), verbose=1)

    # check the residuals
    residual = lopcst.evaluate(newring.radiation_on(copy=True))
    assert_close(residual, 0, rtol=0.0, atol=6e-9)

    # Define the constraints
    lincst = at.LinoptConstraints(test_ring)
    lincst.add('tunes', [0.38, 0.85], name='tunes')
    lopcst = at.EnvelopeConstraints(test_ring)
    lopcst.add('tunes', 1.1e-4, index=2, name='sync. tune')

    # Perform the fit
    newring = at.match(test_ring, variables, (lincst, lopcst), verbose=1)

    # check the residuals
    linresidual = lincst.evaluate(newring)
    lopresidual = lopcst.evaluate(newring.radiation_on(copy=True))
    assert_close(linresidual, 0, rtol=0.0, atol=6e-9)
    assert_close(lopresidual, 0, rtol=0.0, atol=3e-8)
